@@ -2,7 +2,7 @@
 #
 # File: b2/console_tool.py
 #
-# Copyright 2016 Backblaze Inc. All Rights Reserved.
+# Copyright 2018 Backblaze Inc. All Rights Reserved.
 #
 # License https://www.backblaze.com/using_b2_code.html
 #
@@ -10,6 +10,9 @@
 
 from __future__ import absolute_import, print_function
 
+import copy
+import datetime
+import functools
 import getpass
 import json
 import locale
@@ -30,14 +33,15 @@ from .account_info.sqlite_account_info import (
 from .account_info.test_upload_url_concurrency import test_upload_url_concurrency
 from .account_info.exception import (MissingAccountData)
 from .api import (B2Api)
-from .b2http import (test_http, B2Http)
+from .b2http import (test_http)
 from .cache import (AuthInfoCache)
 from .download_dest import (DownloadDestLocalFile)
 from .exception import (B2Error, BadFileInfo)
+from .sync.scan_policies import ScanPoliciesManager
 from .file_version import (FileVersionInfo)
 from .parse_args import parse_arg_list
 from .progress import (make_progress_listener)
-from .raw_api import (SRC_LAST_MODIFIED_MILLIS, B2RawApi, test_raw_api)
+from .raw_api import (SRC_LAST_MODIFIED_MILLIS, test_raw_api)
 from .sync import parse_sync_folder, sync_folders
 from .utils import (current_time_millis, set_shutting_down)
 from .version import (VERSION)
@@ -76,6 +80,20 @@ def mixed_case_to_hyphens(s):
     return s[0].lower() + ''.join(c if c.islower() else '-' + c.lower() for c in s[1:])
 
 
+def parse_comma_separated_list(s):
+    return [word.strip() for word in s.split(',')]
+
+
+def apply_or_none(fcn, value):
+    """
+    If the value is None, return None, otherwise return the result of applying the function to it.
+    """
+    if value is None:
+        return None
+    else:
+        return fcn(value)
+
+
 class Command(object):
     """
     Base class for commands.  Has basic argument parsing and printing.
@@ -190,19 +208,26 @@ class Command(object):
 
 class AuthorizeAccount(Command):
     """
-    b2 authorize-account [<accountId>] [<applicationKey>]
+    b2 authorize-account [<accountIdOrKeyId>] [<applicationKey>]
 
-        Prompts for Backblaze accountID and applicationKey (unless they are given
-        on the command line).
+        Prompts for Backblaze accountID and applicationKey (unless they
+        are given on the command line).
 
-        The account ID is a 12-digit hex number that you can get from
-        your account page on backblaze.com.
+        You can authorize with either the master application key or
+        a normal application key.
 
-        The application key is a 40-digit hex number that you can get from
-        your account page on backblaze.com.
+        To use the master application key, provide the account ID and
+        application key from the "B2 Cloud Storage Buckets" page on
+        the web site: https://secure.backblaze.com/b2_buckets.htm
+
+        To use a normal application key, created with the create-key
+        command or on the web site, provide the application key ID
+        and the application key itself.
 
         Stores an account auth token in {B2_ACCOUNT_INFO_DEFAULT_FILE} by default,
         or the file specified by the {B2_ACCOUNT_INFO_ENV_VAR} environment variable.
+
+        Requires capability: listBuckets
     """
 
     OPTION_FLAGS = ['dev', 'staging']  # undocumented
@@ -231,6 +256,25 @@ class AuthorizeAccount(Command):
 
         try:
             self.api.authorize_account(realm, args.accountId, args.applicationKey)
+
+            allowed = self.api.account_info.get_allowed()
+            if 'listBuckets' not in allowed['capabilities']:
+                logger.error(
+                    'ConsoleTool cannot work with a bucket-restricted key and no listBuckets capability'
+                )
+                self._print_stderr(
+                    'ERROR: application key has no listBuckets capability, which is required for the b2 command-line tool'
+                )
+                self.api.account_info.clear()
+                return 1
+            if allowed['bucketId'] is not None and allowed['bucketName'] is None:
+                logger.error('ConsoleTool has bucket-restricted key and the bucket does not exist')
+                self._print_stderr(
+                    "ERROR: application key is restricted to bucket id '%s', which no longer exists"
+                    % (allowed['bucketId'],)
+                )
+                self.api.account_info.clear()
+                return 1
             return 0
         except B2Error as e:
             logger.exception('ConsoleTool account authorization error')
@@ -245,6 +289,8 @@ class CancelAllUnfinishedLargeFiles(Command):
         Lists all large files that have been started but not
         finsished and cancels them.  Any parts that have been
         uploaded will be deleted.
+
+        Requires capability: writeFiles
     """
 
     REQUIRED = ['bucketName']
@@ -260,6 +306,13 @@ class CancelAllUnfinishedLargeFiles(Command):
 class CancelLargeFile(Command):
     """
     b2 cancel-large-file <fileId>
+
+        Cancels a large file upload.  Used to undo a start-large-file.
+
+        Cannot be used once the file is finished.  After finishing,
+        using delete-file-version to delete the large file.
+
+        Requires capability: writeFiles
     """
 
     REQUIRED = ['fileId']
@@ -291,6 +344,8 @@ class CreateBucket(Command):
 
         Optionally stores bucket info, CORS rules and lifecycle rules with the bucket.
         These can be given as JSON on the command line.
+
+        Requires capability: writeBuckets
     """
 
     REQUIRED = ['bucketName', 'bucketType']
@@ -311,11 +366,65 @@ class CreateBucket(Command):
         return 0
 
 
+class CreateKey(Command):
+    """
+    b2 create-key [--duration <validDurationSeconds>] [--bucket <bucketName>] [--namePrefix <namePrefix>] <keyName> <capabilities>
+
+        Creates a new application key.  Prints the application key information.  This is the only
+        time the application key itself will be returned.  Listing application keys will show
+        their IDs, but not the secret keys.
+
+        The capabilities are passed in as a comma-separated list, like "readFiles,writeFiles".
+
+        The 'validDurationSeconds' is the length of time the new application key will exist.
+        When the time expires the key will disappear and will no longer be usable.  If not
+        specified, the key will not expire.
+
+        The 'bucketName' is the name of a bucket in the account.  When specified, the key
+        will only allow access to that bucket.
+
+        The 'namePrefix' restricts file access to files whose names start with the prefix.
+
+        The output is the new application key ID, followed by the application key itself.
+        The two values returned are the two that you pass to authorize-account to use the key.
+
+        Requires capability: writeKeys
+    """
+
+    REQUIRED = ['keyName', 'capabilities']
+
+    OPTION_ARGS = ['bucket', 'namePrefix', 'duration']
+
+    ARG_PARSER = {'capabilities': parse_comma_separated_list, 'duration': int}
+
+    def run(self, args):
+        # Translate the bucket name into a bucketId
+        if args.bucket is None:
+            bucket_id_or_none = None
+        else:
+            bucket_id_or_none = self.api.get_bucket_by_name(args.bucket).id_
+
+        response = self.api.create_key(
+            capabilities=args.capabilities,
+            key_name=args.keyName,
+            valid_duration_seconds=args.duration,
+            bucket_id=bucket_id_or_none,
+            name_prefix=args.namePrefix
+        )
+
+        application_key_id = response['applicationKeyId']
+        application_key = response['applicationKey']
+        self._print(application_key_id + " " + application_key)
+        return 0
+
+
 class DeleteBucket(Command):
     """
     b2 delete-bucket <bucketName>
 
         Deletes the bucket with the given name.
+
+        Requires capability: deleteBuckets
     """
 
     REQUIRED = ['bucketName']
@@ -337,6 +446,8 @@ class DeleteFileVersion(Command):
         If you omit the fileName, it requires an initial query to B2
         to get the file name, before making the call to delete the
         file.
+
+        Requires capability: deleteFiles
     """
 
     OPTIONAL_BEFORE = ['fileName']
@@ -347,6 +458,7 @@ class DeleteFileVersion(Command):
             file_name = args.fileName
         else:
             file_name = self._get_file_name_from_file_id(args.fileId)
+
         file_info = self.api.delete_file_version(args.fileId, file_name)
         self._print(json.dumps(file_info.as_dict(), indent=2, sort_keys=True))
         return 0
@@ -356,6 +468,23 @@ class DeleteFileVersion(Command):
         return file_info['fileName']
 
 
+class DeleteKey(Command):
+    """
+    b2 delete-key <applicationKeyId>
+
+        Deletes the specified application key by its 'ID'.
+
+        Requires capability: deleteKeys
+    """
+
+    REQUIRED = ['applicationKeyId']
+
+    def run(self, args):
+        response = self.api.delete_key(application_key_id=args.applicationKeyId)
+        self._print(response['applicationKeyId'])
+        return 0
+
+
 class DownloadFileById(Command):
     """
     b2 download-file-by-id [--noProgress] <fileId> <localFileName>
@@ -365,6 +494,8 @@ class DownloadFileById(Command):
         If the 'tqdm' library is installed, progress bar is displayed
         on stderr.  Without it, simple text progress is printed.
         Use '--noProgress' to disable progress reporting.
+
+        Requires capability: readFiles
     """
 
     OPTION_FLAGS = ['noProgress']
@@ -383,6 +514,12 @@ class DownloadFileByName(Command):
     b2 download-file-by-name [--noProgress] <bucketName> <fileName> <localFileName>
 
         Downloads the given file, and stores it in the given local file.
+
+        If the 'tqdm' library is installed, progress bar is displayed
+        on stderr.  Without it, simple text progress is printed.
+        Use '--noProgress' to disable progress reporting.
+
+        Requires capability: readFiles
     """
 
     OPTION_FLAGS = ['noProgress']
@@ -401,13 +538,15 @@ class GetAccountInfo(Command):
     """
     b2 get-account-info
 
-        Shows the account ID, key, auth token, and URLs.
+        Shows the account ID, key, auth token, URLs, and what capabilities
+        the current application keys has.
     """
 
     def run(self, args):
         account_info = self.api.account_info
         data = dict(
             accountId=account_info.get_account_id(),
+            allowed=account_info.get_allowed(),
             applicationKey=account_info.get_application_key(),
             accountAuthToken=account_info.get_account_auth_token(),
             apiUrl=account_info.get_api_url(),
@@ -419,21 +558,54 @@ class GetAccountInfo(Command):
 
 class GetBucket(Command):
     """
-    b2 get-bucket <bucketName>
+    b2 get-bucket [--showSize] <bucketName>
 
         Prints all of the information about the bucket, including
         bucket info, CORS rules and lifecycle rules.
+
+        If --showSize is specified, then display the number of files
+        (fileCount) in the bucket and the aggregate size of all files
+        (totalSize). Hidden files and hide markers are accounted for
+        in the reported number of files, and hidden files also
+        contribute toward the reported aggregate size, whereas hide
+        markers do not. Each version of a file counts as an individual
+        file, and its size contributes toward the aggregate size.
+        Analysis is recursive. Note that --showSize requires multiple
+        API calls, and will therefore incur additional latency,
+        computation, and Class C transactions.
+
+        Requires capability: listBuckets
     """
 
+    OPTION_FLAGS = ['showSize']
+
     REQUIRED = ['bucketName']
 
     def run(self, args):
         # This always wants up-to-date info, so it does not use
         # the bucket cache.
-        for b in self.api.list_buckets():
-            if b.name == args.bucketName:
+        for b in self.api.list_buckets(args.bucketName):
+            if not args.showSize:
                 self._print(json.dumps(b.bucket_dict, indent=4, sort_keys=True))
                 return 0
+            else:
+                # `files` is a generator. We don't want to collect all of the values from the
+                # generator, as there many be billions of files in a large bucket.
+                files = b.ls("", show_versions=True, recursive=True)
+                # `files` yields tuples of (file_version_info, folder_name). We don't care about
+                # `folder_name`, so just access the first slot of the tuple directly in the
+                # reducer. We can't ask a generator for its size, as the elements are yielded
+                # lazily, so we need to accumulate the count as we go. By using a tuple of
+                # (file count, total size), we can obtain the desired information very compactly
+                # and efficiently.
+                count_size_tuple = functools.reduce(
+                    (lambda partial, f: (partial[0] + 1, partial[1] + f[0].size)), files, (0, 0)
+                )
+                result = copy.copy(b.bucket_dict)
+                result['fileCount'] = count_size_tuple[0]
+                result['totalSize'] = count_size_tuple[1]
+                self._print(json.dumps(result, indent=4, sort_keys=True))
+                return 0
         self._print_stderr('bucket not found: ' + args.bucketName)
         return 1
 
@@ -443,6 +615,8 @@ class GetFileInfo(Command):
     b2 get-file-info <fileId>
 
         Prints all of the information about the file, but not its contents.
+
+        Requires capability: readFiles
     """
 
     REQUIRED = ['fileId']
@@ -466,6 +640,8 @@ class GetDownloadAuth(Command):
         Only files that match that given prefix can be downloaded with
         the token.  The prefix defaults to "", which matches all files
         in the bucket.
+
+        Requires capability: shareFiles
     """
 
     OPTION_ARGS = ['prefix', 'duration']
@@ -499,6 +675,8 @@ class GetDownloadUrlWithAuth(Command):
 
         The token is valid for the duration specified, which defaults
         to 86400 seconds (one day).
+
+        Requires capability: shareFiles
     """
 
     OPTION_ARGS = ['duration']
@@ -545,6 +723,8 @@ class HideFile(Command):
     b2 hide-file <bucketName> <fileName>
 
         Uploads a new, hidden, version of the given file.
+
+        Requires capability: writeFiles
     """
 
     REQUIRED = ['bucketName', 'fileName']
@@ -567,6 +747,8 @@ class ListBuckets(Command):
         and look like this:
 
             98c960fd1cb4390c5e0f0519  allPublic   my-bucket
+
+        Requires capability: listBuckets
     """
 
     def run(self, args):
@@ -583,6 +765,8 @@ class ListFileVersions(Command):
         given point.  This is a low-level operation that reports the
         raw JSON returned from the service.  'b2 ls' provides a higher-
         level view.
+
+        Requires capability: listFiles
     """
 
     REQUIRED = ['bucketName']
@@ -604,6 +788,8 @@ class ListFileNames(Command):
 
         Lists the names of the files in a bucket, starting at the
         given point.
+
+        Requires capability: listFiles
     """
 
     REQUIRED = ['bucketName']
@@ -619,6 +805,97 @@ class ListFileNames(Command):
         return 0
 
 
+class ListKeys(Command):
+    """
+    b2 list-keys
+
+       Lists the application keys for the current account.
+
+       The columns in the output are:
+           - ID of the application key
+           - Name of the application key
+           - Name of the bucket the key is restricted to, or '-' for no restriction
+           - Date of expiration, or '-'
+           - Time of expiration, or '-'
+           - File name prefix, in single quotes
+           - Command-separated list of capabilities
+
+        None of the values contain whitespace.
+
+        For keys restricted to buckets that do not exist any more, the bucket name is
+        replaced with 'id=<bucketId>', because deleted buckets do not have names any
+        more.
+
+        Requires capability: listKeys
+    """
+
+    OPTION_FLAGS = ['long']
+
+    def __init__(self, console_tool):
+        super(ListKeys, self).__init__(console_tool)
+        self.bucket_id_to_bucket_name = None
+
+    def run(self, args):
+        # The first query doesn't pass in a starting key id
+        start_id = None
+
+        # Keep querying until there are no more.
+        while True:
+            # Get some keys and print them
+            response = self.api.list_keys(start_id)
+            self.print_keys(response['keys'], args.long)
+
+            # Are there more?  If so, we'll set the start_id for the next time around.
+            next_id = response.get('nextApplicationKeyId')
+            if next_id is None:
+                break
+            else:
+                start_id = next_id
+
+        return 0
+
+    def print_keys(self, keys_from_response, is_long_format):
+        if is_long_format:
+            format_str = "{keyId}   {keyName:20s}   {bucketName:20s}   {dateStr:10s}   {timeStr:8s}   '{namePrefix}'   {capabilities}"
+        else:
+            format_str = '{keyId}   {keyName:20s}'
+        for key in keys_from_response:
+            timestamp_or_none = apply_or_none(int, key.get('expirationTimestamp'))
+            (date_str, time_str) = self.timestamp_display(timestamp_or_none)
+            key_str = format_str.format(
+                keyId=key['applicationKeyId'],
+                keyName=key['keyName'],
+                bucketName=self.bucket_display_name(key.get('bucketId')),
+                namePrefix=(key.get('namePrefix') or ''),
+                capabilities=','.join(key['capabilities']),
+                dateStr=date_str,
+                timeStr=time_str
+            )
+            self._print(key_str)
+
+    def bucket_display_name(self, bucket_id):
+        # Special case for no bucket ID
+        if bucket_id is None:
+            return '-'
+
+        # Make sure we have the map
+        if self.bucket_id_to_bucket_name is None:
+            self.bucket_id_to_bucket_name = dict((b.id_, b.name) for b in self.api.list_buckets())
+
+        return self.bucket_id_to_bucket_name.get(bucket_id, 'id=' + bucket_id)
+
+    def timestamp_display(self, timestamp_or_none):
+        """
+        Returns a pair (date_str, time_str) for the given timestamp
+        """
+        if timestamp_or_none is None:
+            return '-', '-'
+        else:
+            timestamp = timestamp_or_none
+            dt = datetime.datetime.utcfromtimestamp(timestamp / 1000)
+            return dt.strftime('%Y-%m-%d'), dt.strftime('%H:%M:%S')
+
+
 class ListParts(Command):
     """
     b2 list-parts <largeFileId>
@@ -626,6 +903,8 @@ class ListParts(Command):
         Lists all of the parts that have been uploaded for the given
         large file, which must be a file that was started but not
         finished or canceled.
+
+        Requires capability: writeFiles
     """
 
     REQUIRED = ['largeFileId']
@@ -643,6 +922,8 @@ class ListUnfinishedLargeFiles(Command):
         Lists all of the large files in the bucket that were started,
         but not finished or canceled.
 
+        Requires capability: listFiles
+
     """
 
     REQUIRED = ['bucketName']
@@ -663,7 +944,7 @@ class ListUnfinishedLargeFiles(Command):
 
 class Ls(Command):
     """
-    b2 ls [--long] [--versions] <bucketName> [<folderName>]
+    b2 ls [--long] [--versions] [--recursive] <bucketName> [<folderName>]
 
         Using the file naming convention that "/" separates folder
         names from their contents, returns a list of the files
@@ -678,9 +959,14 @@ class Ls(Command):
 
         The --version option shows all of versions of each file, not
         just the most recent.
+
+        The --recursive option will descend into folders, and will show
+        only files, not folders.
+
+        Requires capability: listFiles
     """
 
-    OPTION_FLAGS = ['long', 'versions']
+    OPTION_FLAGS = ['long', 'versions', 'recursive']
 
     REQUIRED = ['bucketName']
 
@@ -695,7 +981,9 @@ class Ls(Command):
                 prefix += '/'
 
         bucket = self.api.get_bucket_by_name(args.bucketName)
-        for file_version_info, folder_name in bucket.ls(prefix, args.versions):
+        for file_version_info, folder_name in bucket.ls(
+            prefix, show_versions=args.versions, recursive=args.recursive
+        ):
             if not args.long:
                 self._print(folder_name or file_version_info.file_name)
             elif folder_name is not None:
@@ -727,6 +1015,7 @@ class Sync(Command):
             [--compareVersions <option>] [--compareThreshold N] \\
             [--threads N] [--noProgress] [--dryRun ] [--allowEmptySource ] \\
             [--excludeRegex <regex> [--includeRegex <regex>]] \\
+            [--excludeDirRegex <regex>] \\
             <source> <destination>
 
         Copies multiple files from source to destination.  Optionally
@@ -767,9 +1056,25 @@ class Sync(Command):
 
         Note that --includeRegex cannot be used without --excludeRegex.
 
+        When a directory is excluded by using --excludeDirRegex, all of
+        the files within it are excluded, even if they match an --includeRegex
+        pattern.   This means that there is no need to look inside excluded
+        directories, and you can exclude directories containing files for which
+        you don't have read permission and avoid getting errors.
+
+        The --excludeDirRegex is a regular expression that is tested against
+        the full path of each directory.  The path being matched does not have
+        a trailing '/', so don't include on in your regular expression.
+
         Multiple regex rules can be applied by supplying them as pipe
         delimitered instructions. Note that the regex for this command
-        is Python regex. Reference: https://docs.python.org/2/library/re.html
+        is Python regex. Reference: https://docs.python.org/2/library/re.html.
+
+        Regular expressions are considered a match if they match a substring
+        starting at the first character.  ".*e" will match "hello".  This is
+        not ideal, but we will maintain this behavior for compatibility.
+        If you want to match the entire path, put a "$" at the end of the
+        regex, such as ".*llo$".
 
         Files are considered to be the same if they have the same name
         and modification time.  This behaviour can be changed using the
@@ -822,6 +1127,8 @@ class Sync(Command):
         and .Spotlight-V100 folders
             b2 sync -excludeRegex '(.*\.DS_Store)|(.*\.Spotlight-V100)' ... b2://...
 
+        Requires capabilities: listFiles, readFiles (for downloading), writeFiles (for uploading)
+
     """
 
     OPTION_FLAGS = [
@@ -829,7 +1136,7 @@ class Sync(Command):
     ]
     OPTION_ARGS = ['keepDays', 'threads', 'compareVersions', 'compareThreshold']
     REQUIRED = ['source', 'destination']
-    LIST_ARGS = ['excludeRegex', 'includeRegex']
+    LIST_ARGS = ['excludeRegex', 'includeRegex', 'excludeDirRegex']
     ARG_PARSER = {'keepDays': float, 'threads': int, 'compareThreshold': int}
 
     def run(self, args):
@@ -845,6 +1152,11 @@ class Sync(Command):
         source = parse_sync_folder(args.source, self.console_tool.api)
         destination = parse_sync_folder(args.destination, self.console_tool.api)
         allow_empty_source = args.allowEmptySource or VERSION_0_COMPATIBILITY
+        policies_manager = ScanPoliciesManager(
+            exclude_dir_regexes=args.excludeDirRegex,
+            exclude_file_regexes=args.excludeRegex,
+            include_file_regexes=args.includeRegex,
+        )
         sync_folders(
             source_folder=source,
             dest_folder=destination,
@@ -853,6 +1165,7 @@ class Sync(Command):
             stdout=self.stdout,
             no_progress=args.noProgress,
             max_workers=max_workers,
+            policies_manager=policies_manager,
             dry_run=args.dryRun,
             allow_empty_source=allow_empty_source
         )
@@ -883,8 +1196,7 @@ class TestRawApi(Command):
     PRIVATE = True
 
     def run(self, args):
-        test_raw_api()
-        return 0
+        return test_raw_api()
 
 
 class TestUploadUrlConcurrency(Command):
@@ -910,6 +1222,8 @@ class UpdateBucket(Command):
 
         Optionally stores bucket info, CORS rules and lifecycle rules with the bucket.
         These can be given as JSON on the command line.
+
+        Requires capability: writeBuckets
     """
 
     REQUIRED = ['bucketName', 'bucketType']
@@ -960,6 +1274,8 @@ class UploadFile(Command):
         Use '--noProgress' to disable progress reporting.
 
         Each fileInfo is of the form "a=b".
+
+        Requires capability: writeFiles
     """
 
     OPTION_FLAGS = ['noProgress', 'quiet']
@@ -969,7 +1285,6 @@ class UploadFile(Command):
     ARG_PARSER = {'minPartSize': int, 'threads': int}
 
     def run(self, args):
-
         file_infos = {}
         for info in args.info:
             parts = info.split('=', 1)
@@ -986,16 +1301,15 @@ class UploadFile(Command):
         self.api.set_thread_pool_size(max_workers)
 
         bucket = self.api.get_bucket_by_name(args.bucketName)
-        with make_progress_listener(args.localFilePath, args.noProgress) as progress_listener:
-            file_info = bucket.upload_local_file(
-                local_file=args.localFilePath,
-                file_name=args.b2FileName,
-                content_type=args.contentType,
-                file_infos=file_infos,
-                sha1_sum=args.sha1,
-                min_part_size=args.minPartSize,
-                progress_listener=progress_listener
-            )
+        file_info = bucket.upload_local_file(
+            local_file=args.localFilePath,
+            file_name=args.b2FileName,
+            content_type=args.contentType,
+            file_infos=file_infos,
+            sha1_sum=args.sha1,
+            min_part_size=args.minPartSize,
+            progress_listener=make_progress_listener(args.localFilePath, args.noProgress),
+        )
         response = file_info.as_dict()
         if not args.quiet:
             self._print("URL by file name: " + bucket.get_download_url(args.b2FileName))
@@ -1120,6 +1434,13 @@ class ConsoleTool(object):
         The default file to use is: {B2_ACCOUNT_INFO_DEFAULT_FILE}
 
         For more details on one command: b2 help <command>
+        
+        When authorizing with application keys, this tool requires that the key
+        have the 'listBuckets' capability so that it can take the bucket names 
+        you provide on the command line and translate them into bucket IDs for the 
+        B2 Storage service.  Each different command may required additional 
+        capabilities.  You can find the details for each command in the help for 
+        that command.
         '''
         self._print_stderr(textwrap.dedent(epilog).format(**DOC_STRING_DATA))
         return 1
@@ -1185,9 +1506,7 @@ def decode_sys_argv():
 
 def main():
     info = SqliteAccountInfo()
-    b2_http = B2Http()
-    raw_api = B2RawApi(b2_http)
-    b2_api = B2Api(info, AuthInfoCache(info), raw_api=raw_api)
+    b2_api = B2Api(info, AuthInfoCache(info))
     ct = ConsoleTool(b2_api=b2_api, stdout=sys.stdout, stderr=sys.stderr)
     decoded_argv = decode_sys_argv()
     exit_status = ct.run_command(decoded_argv)
