@@ -2,7 +2,7 @@
 #
 # File: b2/raw_api.py
 #
-# Copyright 2016 Backblaze Inc. All Rights Reserved.
+# Copyright 2018 Backblaze Inc. All Rights Reserved.
 #
 # License https://www.backblaze.com/using_b2_code.html
 #
@@ -11,21 +11,35 @@
 from __future__ import print_function
 
 import base64
-import hashlib
 import os
 import random
 import re
 import sys
 import time
-from abc import (ABCMeta, abstractmethod)
+import traceback
+from abc import ABCMeta, abstractmethod
 
 import six
 
-from .b2http import (B2Http)
-from .download_dest import DownloadDestBytes
-from .exception import ChecksumMismatch, TruncatedOutput, UnexpectedCloudBehaviour
+from .b2http import B2Http
+from .exception import UnusableFileName
 from .utils import b2_url_encode, hex_sha1_of_stream
 
+# All possible capabilities
+ALL_CAPABILITIES = [
+    'listKeys',
+    'writeKeys',
+    'deleteKeys',
+    'listBuckets',
+    'writeBuckets',
+    'deleteBuckets',
+    'listFiles',
+    'readFiles',
+    'shareFiles',
+    'writeFiles',
+    'deleteFiles',
+]
+
 # Standard names for file info entries
 SRC_LAST_MODIFIED_MILLIS = 'src_last_modified_millis'
 
@@ -100,6 +114,12 @@ class AbstractRawApi(object):
     ):
         pass
 
+    def get_download_url_by_id(self, download_url, account_auth_token, file_id):
+        return download_url + '/b2api/v1/b2_download_file_by_id?fileId=' + file_id
+
+    def get_download_url_by_name(self, download_url, account_auth_token, bucket_name, file_name):
+        return download_url + '/file/' + bucket_name + '/' + b2_url_encode(file_name)
+
 
 class B2RawApi(AbstractRawApi):
     """
@@ -168,6 +188,22 @@ class B2RawApi(AbstractRawApi):
             lifecycleRules=lifecycle_rules
         )
 
+    def create_key(
+        self, api_url, account_auth_token, account_id, capabilities, key_name,
+        valid_duration_seconds, bucket_id, name_prefix
+    ):
+        return self._post_json(
+            api_url,
+            'b2_create_key',
+            account_auth_token,
+            accountId=account_id,
+            capabilities=capabilities,
+            keyName=key_name,
+            validDurationInSeconds=valid_duration_seconds,
+            bucketId=bucket_id,
+            namePrefix=name_prefix,
+        )
+
     def delete_bucket(self, api_url, account_auth_token, account_id, bucket_id):
         return self._post_json(
             api_url,
@@ -186,112 +222,30 @@ class B2RawApi(AbstractRawApi):
             fileName=file_name
         )
 
-    def download_file_by_id(
-        self, download_url, account_auth_token_or_none, file_id, download_dest, range_=None
-    ):
-        url = download_url + '/b2api/v1/b2_download_file_by_id?fileId=' + file_id
-        return self._download_file_from_url(
-            url, account_auth_token_or_none, download_dest, range_=range_
+    def delete_key(self, api_url, account_auth_token, application_key_id):
+        return self._post_json(
+            api_url,
+            'b2_delete_key',
+            account_auth_token,
+            applicationKeyId=application_key_id,
         )
 
-    def download_file_by_name(
-        self,
-        download_url,
-        account_auth_token_or_none,
-        bucket_name,
-        file_name,
-        download_dest,
-        range_=None
-    ):
-        url = download_url + '/file/' + bucket_name + '/' + b2_url_encode(file_name)
-        return self._download_file_from_url(
-            url, account_auth_token_or_none, download_dest, range_=range_
-        )
-
-    def _download_file_from_url(self, url, account_auth_token_or_none, download_dest, range_=None):
+    def download_file_from_url(self, _, account_auth_token_or_none, url, range_=None):
         """
-        Downloads a file from given url and stores it in the given download_destination.
+        Issues a streaming request for download of a file, potentially authorized.
 
-        Returns a dict containing all of the file info from the headers in the reply.
-
-        :param url: The full URL to download from
+        :param _: unused (caused by B2Session magic)
         :param account_auth_token_or_none: an optional account auth token to pass in
-        :param download_dest: where to put the file when it is downloaded
-        :param progress_listener: where to notify about progress downloading
-        :return:
+        :param url: The full URL to download from
+        :param range: two-element tuple for http Range header
+        :return: b2_http response
         """
         request_headers = {}
-        if range_ is not None:
-            assert len(range_) == 2, range_
-            assert (range_[0] + 0) <= (range_[1] + 0), range_  # not strings
-            assert range_[0] >= 0, range_
-            assert range_[1] >= 1, range_
-            request_headers['Range'] = "bytes=%d-%d" % range_
+        _add_range_header(request_headers, range_)
 
         if account_auth_token_or_none is not None:
             request_headers['Authorization'] = account_auth_token_or_none
-
-        with self.b2_http.get_content(url, request_headers) as response:
-
-            info = response.headers
-
-            file_id = info['x-bz-file-id']
-            file_name = info['x-bz-file-name']
-            content_type = info['content-type']
-            content_length = int(info['content-length'])
-            content_sha1 = info['x-bz-content-sha1']
-            if range_ is not None:
-                if 'Content-Range' not in info:
-                    raise UnexpectedCloudBehaviour('Content-Range header was expected')
-            file_info = dict((k[10:], info[k]) for k in info if k.startswith('x-bz-info-'))
-
-            if SRC_LAST_MODIFIED_MILLIS in file_info:
-                mod_time_millis = int(file_info[SRC_LAST_MODIFIED_MILLIS])
-            else:
-                mod_time_millis = int(info['x-bz-upload-timestamp'])
-
-            block_size = 4096
-            digest = hashlib.sha1()
-            bytes_read = 0
-
-            with download_dest.make_file_context(
-                file_id,
-                file_name,
-                content_length,
-                content_type,
-                content_sha1,
-                file_info,
-                mod_time_millis,
-                range_=range_
-            ) as file:
-                for data in response.iter_content(chunk_size=block_size):
-                    file.write(data)
-                    digest.update(data)
-                    bytes_read += len(data)
-
-                if range_ is None:
-                    if bytes_read != int(info['content-length']):
-                        raise TruncatedOutput(bytes_read, content_length)
-
-                    if content_sha1 != 'none' and digest.hexdigest() != content_sha1:
-                        raise ChecksumMismatch(
-                            checksum_type='sha1',
-                            expected=content_length,
-                            actual=digest.hexdigest()
-                        )
-                else:
-                    desired_length = range_[1] - range_[0] + 1
-                    if bytes_read != desired_length:
-                        raise TruncatedOutput(bytes_read, desired_length)
-
-            return dict(
-                fileId=file_id,
-                fileName=file_name,
-                contentType=content_type,
-                contentLength=content_length,
-                contentSha1=content_sha1,
-                fileInfo=file_info
-            )
+        return self.b2_http.get_content(url, request_headers)
 
     def finish_large_file(self, api_url, account_auth_token, file_id, part_sha1_array):
         return self._post_json(
@@ -330,13 +284,22 @@ class B2RawApi(AbstractRawApi):
             api_url, 'b2_hide_file', account_auth_token, bucketId=bucket_id, fileName=file_name
         )
 
-    def list_buckets(self, api_url, account_auth_token, account_id):
+    def list_buckets(
+        self,
+        api_url,
+        account_auth_token,
+        account_id,
+        bucket_id=None,
+        bucket_name=None,
+    ):
         return self._post_json(
             api_url,
             'b2_list_buckets',
             account_auth_token,
             accountId=account_id,
-            bucketTypes=['all']
+            bucketTypes=['all'],
+            bucketId=bucket_id,
+            bucketName=bucket_name,
         )
 
     def list_file_names(
@@ -370,6 +333,23 @@ class B2RawApi(AbstractRawApi):
             maxFileCount=max_file_count
         )
 
+    def list_keys(
+        self,
+        api_url,
+        account_auth_token,
+        account_id,
+        max_key_count=None,
+        start_application_key_id=None
+    ):
+        return self._post_json(
+            api_url,
+            'b2_list_keys',
+            account_auth_token,
+            accountId=account_id,
+            maxKeyCount=max_key_count,
+            startApplicationKeyId=start_application_key_id,
+        )
+
     def list_parts(self, api_url, account_auth_token, file_id, start_part_number, max_part_count):
         return self._post_json(
             api_url,
@@ -440,6 +420,51 @@ class B2RawApi(AbstractRawApi):
             **kwargs
         )
 
+    def unprintable_to_hex(self, string):
+        """Replace unprintable chars in string with a hex representation.
+
+        :param string: An arbitrary string, possibly with unprintable characters.
+        :return The string, with unprintable characters changed to hex (e.g., "\x07")
+        """
+        unprintables_pattern = re.compile(r'[\x00-\x1f]')
+
+        def hexify(match):
+            return r'\x{0:02x}'.format(ord(match.group()))
+
+        return unprintables_pattern.sub(hexify, string)
+
+    def check_b2_filename(self, filename):
+        """
+        Raise an appropriate exception with details if the filename is unusable.
+
+        See https://www.backblaze.com/b2/docs/files.html for the rules.
+
+        :param filename: A proposed filename in unicode.
+        :return: None if the filename is usable.
+        """
+        encoded_name = filename.encode('utf-8')
+        length_in_bytes = len(encoded_name)
+        if length_in_bytes < 1:
+            raise UnusableFileName("Filename must be at least 1 character.")
+        if length_in_bytes > 1024:
+            raise UnusableFileName("Filename is too long (can be at most 1024 bytes).")
+        lowest_unicode_value = ord(min(filename))
+        if lowest_unicode_value < 32:
+            message = u"Filename \"{0}\" contains code {1} (hex {2:02x}), less than 32.".format(
+                self.unprintable_to_hex(filename), lowest_unicode_value, lowest_unicode_value
+            )
+            raise UnusableFileName(message)
+        # No DEL for you.
+        if '\x7f' in filename:
+            raise UnusableFileName("DEL character (0x7f) not allowed.")
+        if filename[0] == '/' or filename[-1] == '/':
+            raise UnusableFileName("Filename may not start or end with '/'.")
+        if '//' in filename:
+            raise UnusableFileName("Filename may not contain \"//\".")
+        long_segment = max([len(segment.encode('utf-8')) for segment in filename.split('/')])
+        if long_segment > 250:
+            raise UnusableFileName("Filename segment too long (maximum 250 bytes in utf-8).")
+
     def upload_file(
         self, upload_url, upload_auth_token, file_name, content_length, content_type, content_sha1,
         file_infos, data_stream
@@ -457,6 +482,8 @@ class B2RawApi(AbstractRawApi):
         :param data_stream: A file like object from which the contents of the file can be read.
         :return:
         """
+        # Raise UnusableFileName if the file_name doesn't meet the rules.
+        self.check_b2_filename(file_name)
         headers = {
             'Authorization': upload_auth_token,
             'Content-Length': str(content_length),
@@ -499,8 +526,8 @@ def test_raw_api():
         raw_api = B2RawApi(B2Http())
         test_raw_api_helper(raw_api)
         return 0
-    except Exception as e:
-        print('Caught exception: %s' % (repr(e),), file=sys.stdout)
+    except Exception:
+        traceback.print_exc(file=sys.stdout)
         return 1
 
 
@@ -516,7 +543,6 @@ def test_raw_api_helper(raw_api):
     this test will break and we'll have to do something about
     it.
     """
-
     account_id = os.environ.get('TEST_ACCOUNT_ID')
     if account_id is None:
         print('TEST_ACCOUNT_ID is not set.', file=sys.stderr)
@@ -534,16 +560,40 @@ def test_raw_api_helper(raw_api):
     api_url = auth_dict['apiUrl']
     download_url = auth_dict['downloadUrl']
 
+    # b2_create_key
+    print('b2_create_key')
+    key_dict = raw_api.create_key(
+        api_url,
+        account_auth_token,
+        account_id,
+        ['readFiles'],
+        'testKey',
+        None,
+        None,
+        None,
+    )
+
+    # b2_list_keys
+    print('b2_list_keys')
+    raw_api.list_keys(api_url, account_auth_token, account_id, 10)
+
+    # b2_delete_key
+    print('b2_delete_key')
+    raw_api.delete_key(api_url, account_auth_token, key_dict['applicationKeyId'])
+
     # b2_create_bucket, with a unique bucket name
     # Include the account ID in the bucket name to be
     # sure it doesn't collide with bucket names from
     # other accounts.
     print('b2_create_bucket')
-    bucket_name = '%s-%d-%d' % (account_id, int(time.time()), random.randint(1000, 9999))
+    bucket_name = 'test-raw-api-%s-%d-%d' % (
+        account_id, int(time.time()), random.randint(1000, 9999)
+    )
     bucket_dict = raw_api.create_bucket(
         api_url, account_auth_token, account_id, bucket_name, 'allPublic'
     )
     bucket_id = bucket_dict['bucketId']
+    first_bucket_revision = bucket_dict['revision']
 
     # b2_list_buckets
     print('b2_list_buckets')
@@ -561,51 +611,58 @@ def test_raw_api_helper(raw_api):
     file_contents = six.b('hello world')
     file_sha1 = hex_sha1_of_stream(six.BytesIO(file_contents), len(file_contents))
     file_dict = raw_api.upload_file(
-        upload_url, upload_auth_token, file_name, len(file_contents), 'text/plain', file_sha1,
-        {'color': 'blue'}, six.BytesIO(file_contents)
+        upload_url,
+        upload_auth_token,
+        file_name,
+        len(file_contents),
+        'text/plain',
+        file_sha1,
+        {'color': 'blue'},
+        six.BytesIO(file_contents),
     )
     file_id = file_dict['fileId']
 
     # b2_download_file_by_id with auth
     print('b2_download_file_by_id (auth)')
-    download_dest = DownloadDestBytes()
-    raw_api.download_file_by_id(download_url, account_auth_token, file_id, download_dest)
-    assert file_contents == download_dest.bytes_io.getvalue()
+    url = raw_api.get_download_url_by_id(download_url, None, file_id)
+    with raw_api.download_file_from_url(None, account_auth_token, url) as response:
+        data = next(response.iter_content(chunk_size=len(file_contents)))
+        assert data == file_contents, data
 
     # b2_download_file_by_id no auth
     print('b2_download_file_by_id (no auth)')
-    download_dest = DownloadDestBytes()
-    raw_api.download_file_by_id(download_url, None, file_id, download_dest)
-    assert file_contents == download_dest.bytes_io.getvalue()
+    url = raw_api.get_download_url_by_id(download_url, None, file_id)
+    with raw_api.download_file_from_url(None, None, url) as response:
+        data = next(response.iter_content(chunk_size=len(file_contents)))
+        assert data == file_contents, data
 
     # b2_download_file_by_name with auth
     print('b2_download_file_by_name (auth)')
-    download_dest = DownloadDestBytes()
-    raw_api.download_file_by_name(
-        download_url, account_auth_token, bucket_name, file_name, download_dest
-    )
-    assert file_contents == download_dest.bytes_io.getvalue()
+    url = raw_api.get_download_url_by_name(download_url, None, bucket_name, file_name)
+    with raw_api.download_file_from_url(None, account_auth_token, url) as response:
+        data = next(response.iter_content(chunk_size=len(file_contents)))
+        assert data == file_contents, data
 
     # b2_download_file_by_name no auth
     print('b2_download_file_by_name (no auth)')
-    download_dest = DownloadDestBytes()
-    raw_api.download_file_by_name(download_url, None, bucket_name, file_name, download_dest)
-    assert file_contents == download_dest.bytes_io.getvalue()
+    url = raw_api.get_download_url_by_name(download_url, None, bucket_name, file_name)
+    with raw_api.download_file_from_url(None, None, url) as response:
+        data = next(response.iter_content(chunk_size=len(file_contents)))
+        assert data == file_contents, data
 
     # b2_get_download_authorization
     print('b2_get_download_authorization')
     download_auth = raw_api.get_download_authorization(
-        api_url, account_auth_token, bucket_id, "prefix", 12345
+        api_url, account_auth_token, bucket_id, file_name[:-2], 12345
     )
     download_auth_token = download_auth['authorizationToken']
 
     # b2_download_file_by_name with download auth
     print('b2_download_file_by_name (download auth)')
-    download_dest = DownloadDestBytes()
-    raw_api.download_file_by_name(
-        download_url, download_auth_token, bucket_name, file_name, download_dest
-    )
-    assert file_contents == download_dest.bytes_io.getvalue()
+    url = raw_api.get_download_url_by_name(download_url, None, bucket_name, file_name)
+    with raw_api.download_file_from_url(None, download_auth_token, url) as response:
+        data = next(response.iter_content(chunk_size=len(file_contents)))
+        assert data == file_contents, data
 
     # b2_list_file_names
     print('b2_list_file_names')
@@ -679,11 +736,9 @@ def test_raw_api_helper(raw_api):
         account_id,
         bucket_id,
         'allPrivate',
-        bucket_info={
-            'color': 'blue'
-        }
+        bucket_info={'color': 'blue'}
     )
-    assert updated_bucket['revision'] == 2
+    assert first_bucket_revision < updated_bucket['revision']
 
     # clean up this test
     _clean_and_delete_bucket(raw_api, api_url, account_auth_token, account_id, bucket_id)
@@ -720,7 +775,7 @@ def _clean_and_delete_bucket(raw_api, api_url, account_auth_token, account_id, b
 def _should_delete_bucket(bucket_name):
     # Bucket names for this test look like: c7b22d0b0ad7-1460060364-5670
     # Other buckets should not be deleted.
-    match = re.match(r'^[a-f0-9]+-([0-9]+)-([0-9]+)', bucket_name)
+    match = re.match(r'^test-raw-api-[a-f0-9]+-([0-9]+)-([0-9]+)', bucket_name)
     if match is None:
         return False
 
@@ -728,3 +783,11 @@ def _should_delete_bucket(bucket_name):
     bucket_time = int(match.group(1))
     now = time.time()
     return bucket_time + 3600 <= now
+
+
+def _add_range_header(headers, range_):
+    if range_ is not None:
+        assert len(range_) == 2, range_
+        assert (range_[0] + 0) <= (range_[1] + 0), range_  # not strings
+        assert range_[0] >= 0, range_
+        headers['Range'] = "bytes=%d-%d" % range_
