@@ -2,23 +2,101 @@
 #
 # File: raw_simulator.py
 #
-# Copyright 2015 Backblaze Inc. All Rights Reserved.
+# Copyright 2018 Backblaze Inc. All Rights Reserved.
 #
 # License https://www.backblaze.com/using_b2_code.html
 #
 ######################################################################
 
 import re
+import time
 
 import six
 from six.moves import range
 
+from .b2http import ResponseContextManager
 from .exception import (
-    BadJson, BadUploadUrl, ChecksumMismatch, Conflict, DuplicateBucketName, FileNotPresent,
-    InvalidAuthToken, MissingPart, NonExistentBucket
+    BadJson,
+    BadUploadUrl,
+    ChecksumMismatch,
+    Conflict,
+    DuplicateBucketName,
+    FileNotPresent,
+    InvalidAuthToken,
+    MissingPart,
+    NonExistentBucket,
+    Unauthorized,
 )
 from .raw_api import AbstractRawApi, HEX_DIGITS_AT_END
-from .utils import b2_url_encode
+from .utils import b2_url_decode, b2_url_encode
+
+ALL_CAPABILITES = [
+    'listKeys',
+    'writeKeys',
+    'deleteKeys',
+    'listBuckets',
+    'writeBuckets',
+    'deleteBuckets',
+    'listFiles',
+    'readFiles',
+    'shareFiles',
+    'writeFiles',
+    'deleteFiles',
+]
+
+
+class KeySimulator(object):
+    """
+    Holds information about one application key, which can be either
+    a master application key, or one created with create_key().
+    """
+
+    def __init__(
+        self, account_id, name, key_id, key, capabilities, expiration_timestamp_or_none,
+        bucket_id_or_none, bucket_name_or_none, name_prefix_or_none
+    ):
+        self.name = name
+        self.account_id = account_id
+        self.key_id = key_id
+        self.key = key
+        self.capabilities = capabilities
+        self.expiration_timestamp_or_none = expiration_timestamp_or_none
+        self.bucket_id_or_none = bucket_id_or_none
+        self.bucket_name_or_none = bucket_name_or_none
+        self.name_prefix_or_none = name_prefix_or_none
+
+    def as_key(self):
+        return dict(
+            accountId=self.account_id,
+            bucketId=self.bucket_id_or_none,
+            applicationKey=self.key,
+            applicationKeyId=self.key_id,
+            capabilities=self.capabilities,
+            expirationTimestamp=self.expiration_timestamp_or_none,
+            keyName=self.name,
+            namePrefix=self.name_prefix_or_none,
+        )
+
+    def as_created_key(self):
+        """
+        Returns the dict returned by b2_create_key.
+
+        This is just like the one for b2_list_keys, but also includes the secret key.
+        """
+        result = self.as_key()
+        result['applicationKey'] = self.key
+        return result
+
+    def get_allowed(self):
+        """
+        Returns the 'allowed' structure to include in the response from b2_authorize_account.
+        """
+        return dict(
+            bucketId=self.bucket_id_or_none,
+            bucketName=self.bucket_name_or_none,
+            capabilities=self.capabilities,
+            namePrefix=self.name_prefix_or_none,
+        )
 
 
 class PartSimulator(object):
@@ -70,8 +148,6 @@ class FileSimulator(object):
         self.data_bytes = data_bytes
         self.upload_timestamp = upload_timestamp
         self.range_ = range_
-        if range_ is not None:
-            self.data_bytes = data_bytes[range_[0]:range_[1]]
 
         if action == 'start':
             self.parts = []
@@ -83,6 +159,20 @@ class FileSimulator(object):
         """
         return (self.name, self.file_id)
 
+    def as_download_headers(self):
+        # range has to be added by the caller!
+        headers = {
+            'content-length': len(self.data_bytes) if self.data_bytes is not None else 0,
+            'content-type': self.content_type,
+            'x-bz-content-sha1': self.content_sha1,
+            'x-bz-upload-timestamp': self.upload_timestamp,
+            'x-bz-file-id': self.file_id,
+            'x-bz-file-name': self.name,
+        }  # yapf: disable
+        for key, value in six.iteritems(self.file_info):
+            headers['x-bz-info-' + key] = value
+        return headers
+
     def as_upload_result(self):
         return dict(
             fileId=self.file_id,
@@ -165,11 +255,6 @@ class FileSimulator(object):
             parts = parts[:max_part_count]
         return dict(parts=parts, nextPartNumber=next_part_number)
 
-    def mod_time_millis(self):
-        if 'src_last_modified_millis' in self.file_info:
-            return int(self.file_info['src_last_modified_millis'])
-        return 0
-
 
 class BucketSimulator(object):
 
@@ -237,11 +322,11 @@ class BucketSimulator(object):
         del self.file_id_to_file[file_id]
         return dict(fileId=file_id, fileName=file_name, uploadTimestamp=file_sim.upload_timestamp)
 
-    def download_file_by_id(self, file_id, download_dest, range_=None):
+    def download_file_by_id(self, file_id, range_=None):
         file_sim = self.file_id_to_file[file_id]
-        self._download_file_sim(download_dest, file_sim, range_=range_)
+        return self._download_file_sim(file_sim, range_=range_)
 
-    def download_file_by_name(self, file_name, download_dest, range_=None):
+    def download_file_by_name(self, file_name, range_=None):
         files = self.list_file_names(file_name, 1)['files']
         if len(files) == 0:
             raise FileNotPresent(file_name)
@@ -249,17 +334,10 @@ class BucketSimulator(object):
         if file_dict['fileName'] != file_name or file_dict['action'] != 'upload':
             raise FileNotPresent(file_name)
         file_sim = self.file_name_and_id_to_file[(file_name, file_dict['fileId'])]
-        self._download_file_sim(download_dest, file_sim, range_=range_)
+        return self._download_file_sim(file_sim, range_=range_)
 
-    def _download_file_sim(self, download_dest, file_sim, range_=None):
-        with download_dest.make_file_context(
-            file_sim.file_id, file_sim.name, file_sim.content_length, file_sim.content_type,
-            file_sim.content_sha1, file_sim.file_info, file_sim.mod_time_millis(), range_
-        ) as f:
-            if range_ is None:
-                f.write(file_sim.data_bytes)
-            else:
-                f.write(file_sim.data_bytes[range_[0]:range_[1]])
+    def _download_file_sim(self, file_sim, range_=None):
+        return ResponseContextManager(FakeResponse(file_sim, range_))
 
     def finish_large_file(self, file_id, part_sha1_array):
         file_sim = self.file_id_to_file[file_id]
@@ -441,16 +519,81 @@ class RawSimulator(AbstractRawApi):
 
     MIN_PART_SIZE = 200
 
+    # This is the maximum duration in seconds that an application key can be valid (1000 days).
+    MAX_DURATION_IN_SECONDS = 86400000
+
     UPLOAD_PART_MATCHER = re.compile('https://upload.example.com/part/([^/]*)')
+    DOWNLOAD_URL_MATCHER = re.compile(
+        DOWNLOAD_URL + '(?:' + '|'.join(
+            (
+                '/b2api/v1/b2_download_file_by_id\?fileId=(?P<file_id>[^/]+)',
+                '/file/(?P<bucket_name>[^/]+)/(?P<file_name>.+)',
+            )
+        ) + ')$'
+    )  # yapf: disable
 
     def __init__(self):
-        self.authorized_accounts = set()
+        # Map from account_id or application_key_id to KeySimulator.
+        # The entry for the account_id is for the master application
+        # key for the account, and the entries with application_key_id
+        # are for keys created b2 createKey().
+        self.key_id_to_key = dict()
+
+        # Map from auth token to the KeySimulator for it.
+        self.auth_token_to_key = dict()
+
+        # Set of auth tokens that have expired
+        self.expired_auth_tokens = set()
+
+        # Counter for generating auth tokens.
+        self.auth_token_counter = 0
+
+        # Counter for generating account IDs an their matching master application keys.
+        self.account_counter = 0
+
         self.bucket_name_to_bucket = dict()
         self.bucket_id_to_bucket = dict()
         self.bucket_id_counter = iter(range(100))
         self.file_id_to_bucket_id = {}
+        self.all_application_keys = []
+        self.app_key_counter = 0
         self.upload_errors = []
 
+    def expire_auth_token(self, auth_token):
+        """
+        Simulate the auth token expiring.
+
+        The next call that tries to use this auth token will get an
+        auth_token_expired error.
+        """
+        assert auth_token in self.auth_token_to_key
+        self.expired_auth_tokens.add(auth_token)
+
+    def create_account(self):
+        """
+        Returns (accountId, masterApplicationKey) for a newly created account.
+        """
+        # Pick the IDs for the account and the key
+        account_id = 'account-%d' % (self.account_counter,)
+        master_key = 'masterKey-%d' % (self.account_counter,)
+        self.account_counter += 1
+
+        # Create the key
+        self.key_id_to_key[account_id] = KeySimulator(
+            account_id=account_id,
+            name='master',
+            key_id=account_id,
+            key=master_key,
+            capabilities=ALL_CAPABILITES,
+            expiration_timestamp_or_none=None,
+            bucket_id_or_none=None,
+            bucket_name_or_none=None,
+            name_prefix_or_none=None,
+        )
+
+        # Return the info
+        return (account_id, master_key)
+
     def set_upload_errors(self, errors):
         """
         Stores a sequence of exceptions to raise on upload.  Each one will
@@ -461,24 +604,27 @@ class RawSimulator(AbstractRawApi):
         self.upload_errors = errors
 
     def authorize_account(self, realm_url, account_id, application_key):
-        assert realm_url == 'http://production.example.com'
-        if application_key != 'good-app-key':
-            raise InvalidAuthToken(
-                'invalid application key: %s' % (application_key,), 'bad_auth_token'
-            )
-        self.authorized_accounts.add(account_id)
+        key_sim = self.key_id_to_key.get(account_id)
+        if key_sim is None:
+            raise InvalidAuthToken('account/key ID not valid', 'unauthorized')
+        if application_key != key_sim.key:
+            raise InvalidAuthToken('secret key is wrong', 'unauthorized')
+        auth_token = 'auth_token_%d' % (self.auth_token_counter,)
+        self.auth_token_counter += 1
+        self.auth_token_to_key[auth_token] = key_sim
         return dict(
-            accountId=account_id,
-            authorizationToken='AUTH:' + account_id,
+            accountId=key_sim.account_id,
+            authorizationToken=auth_token,
             apiUrl=self.API_URL,
             downloadUrl=self.DOWNLOAD_URL,
-            minimumPartSize=self.MIN_PART_SIZE
-        )  # yapf: disable
+            minimumPartSize=self.MIN_PART_SIZE,
+            allowed=key_sim.get_allowed(),
+        )
 
     def cancel_large_file(self, api_url, account_auth_token, file_id):
         bucket_id = self.file_id_to_bucket_id[file_id]
         bucket = self._get_bucket_by_id(bucket_id)
-        self._assert_account_auth(api_url, account_auth_token, bucket.account_id)
+        self._assert_account_auth(api_url, account_auth_token, bucket.account_id, 'writeFiles')
         return bucket.cancel_large_file(file_id)
 
     def create_bucket(
@@ -492,9 +638,9 @@ class RawSimulator(AbstractRawApi):
         cors_rules=None,
         lifecycle_rules=None
     ):
-        if not re.match(r'^[-a-zA-Z]*$', bucket_name):
+        if not re.match(r'^[-a-zA-Z0-9]*$', bucket_name):
             raise BadJson('illegal bucket name: ' + bucket_name)
-        self._assert_account_auth(api_url, account_auth_token, account_id)
+        self._assert_account_auth(api_url, account_auth_token, account_id, 'writeBuckets')
         if bucket_name in self.bucket_name_to_bucket:
             raise DuplicateBucketName(bucket_name)
         bucket_id = 'bucket_' + str(six.next(self.bucket_id_counter))
@@ -506,60 +652,109 @@ class RawSimulator(AbstractRawApi):
         self.bucket_id_to_bucket[bucket_id] = bucket
         return bucket.bucket_dict()
 
+    def create_key(
+        self, api_url, account_auth_token, account_id, capabilities, key_name,
+        valid_duration_seconds, bucket_id, name_prefix
+    ):
+        if not re.match(r'^[A-Za-z0-9-]{1,100}$', key_name):
+            raise BadJson('illegal key name: ' + key_name)
+        if valid_duration_seconds is not None:
+            if valid_duration_seconds < 1 or valid_duration_seconds > self.MAX_DURATION_IN_SECONDS:
+                raise BadJson(
+                    'valid duration must be greater than 0, and less than 1000 days in seconds'
+                )
+        self._assert_account_auth(api_url, account_auth_token, account_id, 'writeKeys')
+
+        if valid_duration_seconds is None:
+            expiration_timestamp_or_none = None
+        else:
+            expiration_timestamp_or_none = int(time.time() + valid_duration_seconds)
+
+        index = self.app_key_counter
+        self.app_key_counter += 1
+        app_key_id = 'appKeyId%d' % (index,)
+        app_key = 'appKey%d' % (index,)
+        if bucket_id is None:
+            bucket_name_or_none = None
+        else:
+            bucket_name_or_none = self._get_bucket_by_id(bucket_id).bucket_name
+        key_sim = KeySimulator(
+            account_id=account_id,
+            name=key_name,
+            key_id=app_key_id,
+            key=app_key,
+            capabilities=capabilities,
+            expiration_timestamp_or_none=expiration_timestamp_or_none,
+            bucket_id_or_none=bucket_id,
+            bucket_name_or_none=bucket_name_or_none,
+            name_prefix_or_none=name_prefix
+        )
+        self.key_id_to_key[app_key_id] = key_sim
+        self.all_application_keys.append(key_sim)
+        return key_sim.as_created_key()
+
     def delete_file_version(self, api_url, account_auth_token, file_id, file_name):
         bucket_id = self.file_id_to_bucket_id[file_id]
         bucket = self._get_bucket_by_id(bucket_id)
-        self._assert_account_auth(api_url, account_auth_token, bucket.account_id)
+        self._assert_account_auth(api_url, account_auth_token, bucket.account_id, 'deleteFiles')
         return bucket.delete_file_version(file_id, file_name)
 
     def delete_bucket(self, api_url, account_auth_token, account_id, bucket_id):
-        self._assert_account_auth(api_url, account_auth_token, account_id)
+        self._assert_account_auth(api_url, account_auth_token, account_id, 'deleteBuckets')
         bucket = self._get_bucket_by_id(bucket_id)
         del self.bucket_name_to_bucket[bucket.bucket_name]
         del self.bucket_id_to_bucket[bucket_id]
         return bucket.bucket_dict()
 
-    def download_file_by_id(
-        self, download_url, account_auth_token_or_none, file_id, download_dest, range_=None
-    ):
+    def download_file_from_url(self, _, account_auth_token_or_none, url, range_=None):
         # TODO: check auth token if bucket is not public
-        bucket_id = self.file_id_to_bucket_id[file_id]
-        bucket = self._get_bucket_by_id(bucket_id)
-        bucket.download_file_by_id(file_id, download_dest, range_=range_)
+        matcher = self.DOWNLOAD_URL_MATCHER.match(url)
+        assert matcher is not None, url
+        groupdict = matcher.groupdict()
+        file_id = groupdict['file_id']
+        bucket_name = groupdict['bucket_name']
+        file_name = groupdict['file_name']
+        if file_id is not None:
+            bucket_id = self.file_id_to_bucket_id[file_id]
+            bucket = self._get_bucket_by_id(bucket_id)
+            return bucket.download_file_by_id(file_id, range_=range_)
+        elif bucket_name is not None and file_name is not None:
+            bucket = self._get_bucket_by_name(bucket_name)
+            return bucket.download_file_by_name(b2_url_decode(file_name), range_=range_)
+        else:
+            assert False
 
-    def download_file_by_name(
-        self,
-        download_url,
-        account_auth_token_or_none,
-        bucket_name,
-        file_name,
-        download_dest,
-        range_=None
-    ):
-        assert download_url == self.DOWNLOAD_URL
-        # TODO: check auth token if bucket is not public
-        bucket = self._get_bucket_by_name(bucket_name)
-        bucket.download_file_by_name(file_name, download_dest)
+    def delete_key(self, api_url, account_auth_token, application_key_id):
+        assert api_url == self.API_URL
+        return dict(
+            accountId='accountId',
+            applicationKeyId=application_key_id,
+            keyName='keyName',
+            capabilities=['listBuckets', 'readBuckets', 'writeBuckets']
+        )
 
     def finish_large_file(self, api_url, account_auth_token, file_id, part_sha1_array):
         bucket_id = self.file_id_to_bucket_id[file_id]
         bucket = self._get_bucket_by_id(bucket_id)
-        self._assert_account_auth(api_url, account_auth_token, bucket.account_id)
+        self._assert_account_auth(api_url, account_auth_token, bucket.account_id, 'writeFiles')
         return bucket.finish_large_file(file_id, part_sha1_array)
 
     def get_download_authorization(
         self, api_url, account_auth_token, bucket_id, file_name_prefix, valid_duration_in_seconds
     ):
         bucket = self._get_bucket_by_id(bucket_id)
-        self._assert_account_auth(api_url, account_auth_token, bucket.account_id)
+        self._assert_account_auth(api_url, account_auth_token, bucket.account_id, 'shareFiles')
         return {
             'bucketId':
-            bucket_id,
+                bucket_id,
             'fileNamePrefix':
-            file_name_prefix,
+                file_name_prefix,
             'authorizationToken':
-            'fake_download_auth_token_%s_%s_%d' %
-            (bucket_id, b2_url_encode(file_name_prefix), valid_duration_in_seconds)
+                'fake_download_auth_token_%s_%s_%d' % (
+                    bucket_id,
+                    b2_url_encode(file_name_prefix),
+                    valid_duration_in_seconds,
+                )
         }
 
     def get_file_info(self, api_url, account_auth_token, file_id):
@@ -569,36 +764,67 @@ class RawSimulator(AbstractRawApi):
 
     def get_upload_url(self, api_url, account_auth_token, bucket_id):
         bucket = self._get_bucket_by_id(bucket_id)
-        self._assert_account_auth(api_url, account_auth_token, bucket.account_id)
+        self._assert_account_auth(api_url, account_auth_token, bucket.account_id, 'writeFiles')
         return self._get_bucket_by_id(bucket_id).get_upload_url()
 
     def get_upload_part_url(self, api_url, account_auth_token, file_id):
         bucket_id = self.file_id_to_bucket_id[file_id]
         bucket = self._get_bucket_by_id(bucket_id)
-        self._assert_account_auth(api_url, account_auth_token, bucket.account_id)
+        self._assert_account_auth(api_url, account_auth_token, bucket.account_id, 'writeFiles')
         return self._get_bucket_by_id(bucket_id).get_upload_part_url(file_id)
 
     def hide_file(self, api_url, account_auth_token, bucket_id, file_name):
         bucket = self._get_bucket_by_id(bucket_id)
-        self._assert_account_auth(api_url, account_auth_token, bucket.account_id)
+        self._assert_account_auth(api_url, account_auth_token, bucket.account_id, 'writeFiles')
         response = bucket.hide_file(file_name)
         self.file_id_to_bucket_id[response['fileId']] = bucket_id
         return response
 
-    def list_buckets(self, api_url, account_auth_token, account_id):
-        self._assert_account_auth(api_url, account_auth_token, account_id)
+    def list_buckets(
+        self, api_url, account_auth_token, account_id, bucket_id=None, bucket_name=None
+    ):
+        # First, map the bucket name to a bucket_id, so that we can check auth.
+        if bucket_name is None:
+            bucket_id_for_auth = bucket_id
+        else:
+            bucket_id_for_auth = self._get_bucket_id_or_none_for_bucket_name(bucket_name)
+        self._assert_account_auth(
+            api_url, account_auth_token, account_id, 'listBuckets', bucket_id_for_auth
+        )
+
+        # Do the query
         sorted_buckets = [
-            self.bucket_name_to_bucket[bucket_name]
-            for bucket_name in sorted(six.iterkeys(self.bucket_name_to_bucket))
+            self.bucket_name_to_bucket[name]
+            for name in sorted(six.iterkeys(self.bucket_name_to_bucket))
+        ]
+        bucket_list = [
+            bucket.bucket_dict()
+            for bucket in sorted_buckets if self._bucket_matches(bucket, bucket_id, bucket_name)
         ]
-        bucket_list = [bucket.bucket_dict() for bucket in sorted_buckets]
         return dict(buckets=bucket_list)
 
+    def _get_bucket_id_or_none_for_bucket_name(self, bucket_name):
+        for bucket in six.itervalues(self.bucket_name_to_bucket):
+            if bucket.bucket_name == bucket_name:
+                return bucket.bucket_id
+
+    def _bucket_matches(self, bucket, bucket_id, bucket_name):
+        return (
+            (bucket_id is None or bucket.bucket_id == bucket_id) and
+            (bucket_name is None or bucket.bucket_name == bucket_name)
+        )
+
     def list_file_names(
         self, api_url, account_auth, bucket_id, start_file_name=None, max_file_count=None
     ):
         bucket = self._get_bucket_by_id(bucket_id)
-        self._assert_account_auth(api_url, account_auth, bucket.account_id)
+        self._assert_account_auth(
+            api_url,
+            account_auth,
+            bucket.account_id,
+            'listFiles',
+            bucket_id=bucket_id,
+        )
         return bucket.list_file_names(start_file_name, max_file_count)
 
     def list_file_versions(
@@ -611,20 +837,38 @@ class RawSimulator(AbstractRawApi):
         max_file_count=None
     ):
         bucket = self._get_bucket_by_id(bucket_id)
-        self._assert_account_auth(api_url, account_auth, bucket.account_id)
+        self._assert_account_auth(
+            api_url,
+            account_auth,
+            bucket.account_id,
+            'listFiles',
+            bucket_id=bucket_id,
+        )
         return bucket.list_file_versions(start_file_name, start_file_id, max_file_count)
 
+    def list_keys(
+        self,
+        api_url,
+        account_auth_token,
+        account_id,
+        max_key_count=1000,
+        start_application_key_id=None
+    ):
+        self._assert_account_auth(api_url, account_auth_token, account_id, 'listKeys')
+        keys = map(lambda key: key.as_key(), self.all_application_keys)
+        return dict(keys=keys, nextKeyId=None)
+
     def list_parts(self, api_url, account_auth_token, file_id, start_part_number, max_part_count):
         bucket_id = self.file_id_to_bucket_id[file_id]
         bucket = self._get_bucket_by_id(bucket_id)
-        self._assert_account_auth(api_url, account_auth_token, bucket.account_id)
+        self._assert_account_auth(api_url, account_auth_token, bucket.account_id, 'writeFiles')
         return bucket.list_parts(file_id, start_part_number, max_part_count)
 
     def list_unfinished_large_files(
         self, api_url, account_auth, bucket_id, start_file_id=None, max_file_count=None
     ):
         bucket = self._get_bucket_by_id(bucket_id)
-        self._assert_account_auth(api_url, account_auth, bucket.account_id)
+        self._assert_account_auth(api_url, account_auth, bucket.account_id, 'listFiles')
         start_file_id = start_file_id or ''
         max_file_count = max_file_count or 100
         return bucket.list_unfinished_large_files(start_file_id, max_file_count)
@@ -633,7 +877,7 @@ class RawSimulator(AbstractRawApi):
         self, api_url, account_auth_token, bucket_id, file_name, content_type, file_info
     ):
         bucket = self._get_bucket_by_id(bucket_id)
-        self._assert_account_auth(api_url, account_auth_token, bucket.account_id)
+        self._assert_account_auth(api_url, account_auth_token, bucket.account_id, 'writeFiles')
         result = bucket.start_large_file(file_name, content_type, file_info)
         self.file_id_to_bucket_id[result['fileId']] = bucket_id
         return result
@@ -652,7 +896,7 @@ class RawSimulator(AbstractRawApi):
     ):
         assert bucket_type or bucket_info
         bucket = self._get_bucket_by_id(bucket_id)
-        self._assert_account_auth(api_url, account_auth_token, bucket.account_id)
+        self._assert_account_auth(api_url, account_auth_token, bucket.account_id, 'writeBuckets')
         return bucket.update_bucket(
             bucket_type=bucket_type,
             bucket_info=bucket_info,
@@ -692,10 +936,22 @@ class RawSimulator(AbstractRawApi):
         bucket = self._get_bucket_by_id(bucket_id)
         return bucket.upload_part(file_id, part_number, content_length, sha1_sum, input_stream)
 
-    def _assert_account_auth(self, api_url, account_auth_token, account_id):
+    def _assert_account_auth(
+        self, api_url, account_auth_token, account_id, capability, bucket_id=None, file_name=None
+    ):
+        key_sim = self.auth_token_to_key.get(account_auth_token)
+        assert key_sim is not None
         assert api_url == self.API_URL
-        assert account_auth_token == 'AUTH:' + account_id
-        assert account_id in self.authorized_accounts
+        assert account_id == key_sim.account_id
+        if account_auth_token in self.expired_auth_tokens:
+            raise InvalidAuthToken('auth token expired', 'auth_token_expired')
+        if capability not in key_sim.capabilities:
+            raise Unauthorized('', 'unauthorized')
+        if key_sim.bucket_id_or_none is not None and key_sim.bucket_id_or_none != bucket_id:
+            raise Unauthorized('', 'unauthorized')
+        if key_sim.name_prefix_or_none is not None:
+            if file_name is not None and not file_name.startswith(key_sim.name_prefix_or_none):
+                raise Unauthorized('', 'unauthorized')
 
     def _get_bucket_by_id(self, bucket_id):
         if bucket_id not in self.bucket_id_to_bucket:
@@ -706,3 +962,21 @@ class RawSimulator(AbstractRawApi):
         if bucket_name not in self.bucket_name_to_bucket:
             raise NonExistentBucket(bucket_name)
         return self.bucket_name_to_bucket[bucket_name]
+
+
+class FakeResponse(object):
+    def __init__(self, file_sim, range_=None):
+        self.data_bytes = file_sim.data_bytes
+        self.headers = file_sim.as_download_headers()
+        if range_ is not None:
+            self.data_bytes = self.data_bytes[range_[0]:range_[1] + 1]
+            self.headers['Content-Range'] = '%s-%s' % range_
+
+    def iter_content(self, chunk_size=1):
+        start = 0
+        while start <= len(self.data_bytes):
+            yield self.data_bytes[start:start + chunk_size]
+            start += chunk_size
+
+    def close(self):
+        pass
