@@ -2,7 +2,7 @@
 #
 # File: test_bucket.py
 #
-# Copyright 2015 Backblaze Inc. All Rights Reserved.
+# Copyright 2018 Backblaze Inc. All Rights Reserved.
 #
 # License https://www.backblaze.com/using_b2_code.html
 #
@@ -56,7 +56,7 @@ class StubProgressListener(AbstractProgressListener):
         return ' '.join(self.history)
 
     def set_total_bytes(self, total_byte_count):
-        assert len(self.history) == 0
+        assert len(self.history) == 0, self.history
         self.history.append('%d:' % (total_byte_count,))
 
     def bytes_completed(self, byte_count):
@@ -65,12 +65,6 @@ class StubProgressListener(AbstractProgressListener):
     def close(self):
         self.history.append('closed')
 
-    def __enter__(self):
-        return self
-
-    def __exit__(self, exc_type, exc_val, exc_tb):
-        pass
-
 
 class CanRetry(B2Error):
     """
@@ -91,7 +85,8 @@ class TestCaseWithBucket(TestBase):
         self.simulator = RawSimulator()
         self.account_info = StubAccountInfo()
         self.api = B2Api(self.account_info, raw_api=self.simulator)
-        self.api.authorize_account('production', 'my-account', 'good-app-key')
+        (self.account_id, self.master_key) = self.simulator.create_account()
+        self.api.authorize_account('production', self.account_id, self.master_key)
         self.api_url = self.account_info.get_api_url()
         self.account_auth_token = self.account_info.get_account_auth_token()
         self.bucket = self.api.create_bucket('my-bucket', 'allPublic')
@@ -396,9 +391,7 @@ class TestUpload(TestCaseWithBucket):
         self._upload_part(large_file_id, 1, data[:part_size])
         progress_listener = StubProgressListener()
         file_info = self.bucket.upload_bytes(
-            data, 'file1', progress_listener=progress_listener, file_infos={
-                'property': 'value1'
-            }
+            data, 'file1', progress_listener=progress_listener, file_infos={'property': 'value1'}
         )
         self.assertEqual(large_file_id, file_info.id_)
         self._check_file_contents('file1', data)
@@ -411,9 +404,7 @@ class TestUpload(TestCaseWithBucket):
         self._upload_part(large_file_id, 1, data[:part_size])
         progress_listener = StubProgressListener()
         file_info = self.bucket.upload_bytes(
-            data, 'file1', progress_listener=progress_listener, file_infos={
-                'property': 'value2'
-            }
+            data, 'file1', progress_listener=progress_listener, file_infos={'property': 'value2'}
         )
         self.assertNotEqual(large_file_id, file_info.id_)
         self._check_file_contents('file1', data)
@@ -492,5 +483,5 @@ class TestPartialDownload(TestCaseWithBucket):
         download = DownloadDestBytes()
         progress_listener = StubProgressListener()
         self.bucket.download_file_by_id(file_info.id_, download, progress_listener, range_=(3, 9))
-        self.assertEqual("6: 6 closed", progress_listener.get_history())
-        assert download.get_bytes_written() == six.b('lo wor'), download.get_bytes_written()
+        self.assertEqual("7: 7 closed", progress_listener.get_history())
+        assert download.get_bytes_written() == six.b('lo worl'), download.get_bytes_written()
