@@ -2,7 +2,7 @@
 #
 # File: b2/test_sync.py
 #
-# Copyright 2016 Backblaze Inc. All Rights Reserved.
+# Copyright 2018 Backblaze Inc. All Rights Reserved.
 #
 # License https://www.backblaze.com/using_b2_code.html
 #
@@ -23,6 +23,7 @@ from b2.exception import CommandError, DestFileNewer
 from b2.file_version import FileVersionInfo
 from b2.sync.folder import AbstractFolder, B2Folder, LocalFolder
 from b2.sync.file import File, FileVersion
+from b2.sync.scan_policies import ScanPoliciesManager, DEFAULT_SCAN_MANAGER
 from b2.sync.sync import BoundedQueueExecutor, make_folder_sync_actions, zip_folders
 from b2.sync.folder_parser import parse_sync_folder
 from b2.utils import TempDir
@@ -62,7 +63,13 @@ class TestLocalFolder(TestSync):
         six.u('hello/a/2'),
         six.u('hello/b'),
         six.u('hello0'),
-        six.u('\u81ea\u7531')
+        six.u('inner/a.bin'),
+        six.u('inner/a.txt'),
+        six.u('inner/b.bin'),
+        six.u('inner/b.txt'),
+        six.u('inner/more/a.bin'),
+        six.u('inner/more/a.txt'),
+        six.u('\u81ea\u7531'),
     ]
 
     @classmethod
@@ -100,10 +107,137 @@ class TestLocalFolder(TestSync):
     def test_invalid_permissions(self):
         with TempDir() as tmpdir:
             folder = self._prepare_folder(tmpdir, invalid_permissions=True)
-            self.assertEqual(self.NAMES[1:], list(f.name for f in folder.all_files(self.reporter)))
-            self.reporter.local_permission_error.assert_called_once_with(
-                os.path.join(tmpdir, self.NAMES[0])
+            # tests differ depending on the user running them. "root" will
+            # succeed in os.access(path, os.R_OK) even if the permissions of
+            # the file are 0 as implemented on self._prepare_folder().
+            # use-case: running test suite inside a docker container
+            if not os.access(os.path.join(tmpdir, self.NAMES[0]), os.R_OK):
+                self.assertEqual(
+                    self.NAMES[1:], list(f.name for f in folder.all_files(self.reporter))
+                )
+                self.reporter.local_permission_error.assert_called_once_with(
+                    os.path.join(tmpdir, self.NAMES[0])
+                )
+            else:
+                self.assertEqual(self.NAMES, list(f.name for f in folder.all_files(self.reporter)))
+
+    def _check_file_filters_results(self, policies_manager, expected_scan_results):
+        with TempDir() as tmpdir:
+            folder = self._prepare_folder(tmpdir)
+            self.assertEqual(
+                expected_scan_results,
+                list(f.name for f in folder.all_files(self.reporter, policies_manager))
             )
+            self.reporter.local_access_error.assert_not_called()
+
+    def test_exclusions(self):
+        expected_list = [
+            six.u('.dot_file'),
+            six.u('hello.'),
+            six.u('hello/a/1'),
+            six.u('hello/a/2'),
+            six.u('hello/b'),
+            six.u('hello0'),
+            six.u('inner/a.txt'),
+            six.u('inner/b.txt'),
+            six.u('inner/more/a.txt'),
+            six.u('\u81ea\u7531'),
+        ]
+        polices_manager = ScanPoliciesManager(exclude_file_regexes=['.*\\.bin'])
+        self._check_file_filters_results(polices_manager, expected_list)
+
+    def test_exclude_all(self):
+        expected_list = []
+        polices_manager = ScanPoliciesManager(exclude_file_regexes=['.*'])
+        self._check_file_filters_results(polices_manager, expected_list)
+
+    def test_exclusions_inclusions(self):
+        expected_list = [
+            six.u('.dot_file'),
+            six.u('hello.'),
+            six.u('hello/a/1'),
+            six.u('hello/a/2'),
+            six.u('hello/b'),
+            six.u('hello0'),
+            six.u('inner/a.bin'),
+            six.u('inner/a.txt'),
+            six.u('inner/b.txt'),
+            six.u('inner/more/a.bin'),
+            six.u('inner/more/a.txt'),
+            six.u('\u81ea\u7531'),
+        ]
+        polices_manager = ScanPoliciesManager(
+            exclude_file_regexes=['.*\\.bin'],
+            include_file_regexes=['.*a\\.bin'],
+        )
+        self._check_file_filters_results(polices_manager, expected_list)
+
+    def test_exclude_matches_prefix(self):
+        expected_list = [
+            six.u('.dot_file'),
+            six.u('hello.'),
+            six.u('hello/b'),
+            six.u('hello0'),
+            six.u('inner/b.bin'),
+            six.u('inner/b.txt'),
+            six.u('\u81ea\u7531'),
+        ]
+        polices_manager = ScanPoliciesManager(exclude_file_regexes=['.*a'])
+        self._check_file_filters_results(polices_manager, expected_list)
+
+    def test_exclude_directory(self):
+        expected_list = [
+            six.u('.dot_file'),
+            six.u('hello.'),
+            six.u('hello0'),
+            six.u('\u81ea\u7531'),
+        ]
+        polices_manager = ScanPoliciesManager(
+            exclude_dir_regexes=['hello', 'more', 'hello0'], exclude_file_regexes=['inner']
+        )
+        self._check_file_filters_results(polices_manager, expected_list)
+
+    def test_exclude_directory2(self):
+        expected_list = [
+            six.u('.dot_file'),
+            six.u('hello.'),
+            six.u('hello0'),
+            six.u('\u81ea\u7531'),
+        ]
+        polices_manager = ScanPoliciesManager(exclude_dir_regexes=['hello$', 'inner'],)
+        self._check_file_filters_results(polices_manager, expected_list)
+
+    def test_exclude_directory_trailing_slash_does_not_match(self):
+        expected_list = [
+            six.u('.dot_file'),
+            six.u('hello.'),
+            six.u('hello0'),
+            six.u('inner/a.bin'),
+            six.u('inner/a.txt'),
+            six.u('inner/b.bin'),
+            six.u('inner/b.txt'),
+            six.u('\u81ea\u7531'),
+        ]
+        polices_manager = ScanPoliciesManager(exclude_dir_regexes=['hello$', 'inner/'],)
+        self._check_file_filters_results(polices_manager, expected_list)
+
+    def test_exclusion_with_exact_match(self):
+        expected_list = [
+            six.u('.dot_file'),
+            six.u('hello.'),
+            six.u('hello/a/1'),
+            six.u('hello/a/2'),
+            six.u('hello/b'),
+            six.u('inner/a.bin'),
+            six.u('inner/a.txt'),
+            six.u('inner/b.bin'),
+            six.u('inner/b.txt'),
+            six.u('inner/more/a.bin'),
+            six.u('inner/more/a.txt'),
+            six.u('\u81ea\u7531'),
+        ]
+        polices_manager = ScanPoliciesManager(exclude_file_regexes=['hello0'],)
+        self._check_file_filters_results(polices_manager, expected_list)
 
 
 class TestB2Folder(TestSync):
@@ -161,8 +295,15 @@ class FakeFolder(AbstractFolder):
         self.f_type = f_type
         self.files = files
 
-    def all_files(self, reporter):
-        return iter(self.files)
+    def all_files(self, reporter, policies_manager=DEFAULT_SCAN_MANAGER):
+        for single_file in self.files:
+            if single_file.name.endswith('/'):
+                if policies_manager.should_exclude_directory(single_file.name):
+                    continue
+            else:
+                if policies_manager.should_exclude_file(single_file.name):
+                    continue
+            yield single_file
 
     def folder_type(self):
         return self.f_type
@@ -248,7 +389,7 @@ class TestZipFolders(TestSync):
         folder_a.all_files = MagicMock(return_value=iter([]))
         folder_b.all_files = MagicMock(return_value=iter([]))
         self.assertEqual([], list(zip_folders(folder_a, folder_b, self.reporter)))
-        folder_a.all_files.assert_called_once_with(self.reporter)
+        folder_a.all_files.assert_called_once_with(self.reporter, DEFAULT_SCAN_MANAGER)
         folder_b.all_files.assert_called_once_with(self.reporter)
 
 
@@ -266,6 +407,7 @@ class FakeArgs(object):
         compareVersions=None,
         compareThreshold=None,
         excludeRegex=None,
+        excludeDirRegex=None,
         includeRegex=None,
         debugLogs=True,
         dryRun=False,
@@ -283,6 +425,9 @@ class FakeArgs(object):
         if includeRegex is None:
             includeRegex = []
         self.includeRegex = includeRegex
+        if excludeDirRegex is None:
+            excludeDirRegex = []
+        self.excludeDirRegex = excludeDirRegex
         self.debugLogs = debugLogs
         self.dryRun = dryRun
         self.allowEmptySource = allowEmptySource
@@ -311,8 +456,8 @@ def b2_file(name, mod_times, size=10):
     """
     versions = [
         FileVersion(
-            'id_%s_%d' % (name[0], abs(mod_time)), 'folder/' + name,
-            abs(mod_time), 'upload' if 0 < mod_time else 'hide', size
+            'id_%s_%d' % (name[0], abs(mod_time)), 'folder/' + name, abs(mod_time), 'upload'
+            if 0 < mod_time else 'hide', size
         ) for mod_time in mod_times
     ]  # yapf disable
     return File(name, versions)
@@ -334,6 +479,8 @@ class TestExclusions(TestSync):
         # only local
         file_a = local_file('a.txt', [100])
         file_b = local_file('b.txt', [100])
+        file_d = local_file('d/d.txt', [100])
+        file_e = local_file('e/e.incl', [100])
 
         # both local and remote
         file_bi = local_file('b.txt.incl', [100])
@@ -342,40 +489,39 @@ class TestExclusions(TestSync):
         # only remote
         file_c = local_file('c.txt', [100])
 
-        local_folder = FakeFolder('local', [file_a, file_b, file_bi, file_z])
+        local_folder = FakeFolder('local', [file_a, file_b, file_d, file_e, file_bi, file_z])
         b2_folder = FakeFolder('b2', [file_bi, file_c, file_z])
 
+        policies_manager = ScanPoliciesManager(
+            exclude_dir_regexes=fakeargs.excludeDirRegex,
+            exclude_file_regexes=fakeargs.excludeRegex,
+            include_file_regexes=fakeargs.includeRegex,
+        )
         actions = list(
-            make_folder_sync_actions(local_folder, b2_folder, fakeargs, TODAY, self.reporter)
+            make_folder_sync_actions(
+                local_folder, b2_folder, fakeargs, TODAY, self.reporter, policies_manager
+            )
         )
         self.assertEqual(expected_actions, [str(a) for a in actions])
 
-    def test_file_exclusions(self):
-        expected_actions = [
-            'b2_upload(/dir/a.txt, folder/a.txt, 100)',
-        ]
-        self._check_folder_sync(expected_actions, FakeArgs(excludeRegex=["b\\.txt"]))
-
     def test_file_exclusions_with_delete(self):
         expected_actions = [
             'b2_upload(/dir/a.txt, folder/a.txt, 100)',
             'b2_delete(folder/b.txt.incl, /dir/b.txt.incl, )',
             'b2_delete(folder/c.txt, /dir/c.txt, )',
+            'b2_upload(/dir/d/d.txt, folder/d/d.txt, 100)',
+            'b2_upload(/dir/e/e.incl, folder/e/e.incl, 100)',
         ]
         self._check_folder_sync(expected_actions, FakeArgs(delete=True, excludeRegex=["b\\.txt"]))
 
-    def test_file_exclusions_inclusions(self):
-        expected_actions = [
-            'b2_upload(/dir/a.txt, folder/a.txt, 100)',
-            'b2_upload(/dir/b.txt, folder/b.txt, 100)',
-        ]
-        fakeargs = FakeArgs(excludeRegex=["b\\.txt"], includeRegex=["b\\.txt"])
-        self._check_folder_sync(expected_actions, fakeargs)
-
     def test_file_exclusions_inclusions_with_delete(self):
         expected_actions = [
             'b2_upload(/dir/a.txt, folder/a.txt, 100)',
+            'b2_delete(folder/b.txt.incl, /dir/b.txt.incl, )',
             'b2_delete(folder/c.txt, /dir/c.txt, )',
+            'b2_upload(/dir/d/d.txt, folder/d/d.txt, 100)',
+            'b2_upload(/dir/e/e.incl, folder/e/e.incl, 100)',
+            'b2_upload(/dir/b.txt.incl, folder/b.txt.incl, 100)',
         ]
         fakeargs = FakeArgs(delete=True, excludeRegex=["b\\.txt"], includeRegex=[".*\\.incl"])
         self._check_folder_sync(expected_actions, fakeargs)
@@ -446,6 +592,11 @@ class TestMakeSyncActions(TestSync):
         actions = ['b2_delete(folder/a.txt, id_a_100, )']
         self._check_local_to_b2(None, dst_file, FakeArgs(delete=True), actions)
 
+    def test_delete_large_b2(self):
+        dst_file = b2_file('a.txt', [100])
+        actions = ['b2_delete(folder/a.txt, id_a_100, )']
+        self._check_local_to_b2(None, dst_file, FakeArgs(delete=True), actions)
+
     def test_delete_b2_multiple_versions(self):
         dst_file = b2_file('a.txt', [100, 200])
         actions = [
