@@ -3,7 +3,7 @@
 #
 # File: test_b2_command_line.py
 #
-# Copyright 2015 Backblaze Inc. All Rights Reserved.
+# Copyright 2018 Backblaze Inc. All Rights Reserved.
 #
 # License https://www.backblaze.com/using_b2_code.html
 #
@@ -13,6 +13,7 @@ from __future__ import print_function
 import hashlib
 import json
 import os.path
+import platform
 import random
 import re
 import shutil
@@ -23,14 +24,15 @@ import tempfile
 import threading
 import unittest
 
-from b2.console_tool import VERSION_0_COMPATIBILITY
+from b2.utils import fix_windows_path_limit
 
 USAGE = """
 This program tests the B2 command-line client.
 
 Usages:
 
-    {command} <accountId> <applicationKey> [basic | sync]
+    {command} <accountId> <applicationKey> [basic | sync_down | sync_up | sync_up_no_prefix 
+                                               keys | sync_long_path | download | account]
 
         The optional last argument specifies which of the tests to run.  If not
         specified, all test will run.  Runs the b2 package in the current directory.
@@ -62,7 +64,7 @@ def write_file(path, contents):
 
 
 def file_mod_time_millis(path):
-    return int(round(1000 * os.path.getmtime(path)))
+    return int(os.path.getmtime(path) * 1000)
 
 
 def set_file_mod_time_millis(path, time):
@@ -85,7 +87,7 @@ class TempDir(object):
         return self.dirpath
 
     def __exit__(self, exc_type, exc_val, exc_tb):
-        shutil.rmtree(self.dirpath)
+        shutil.rmtree(fix_windows_path_limit(self.dirpath))
 
 
 class StringReader(object):
@@ -104,8 +106,8 @@ class StringReader(object):
 
 
 def remove_insecure_platform_warnings(text):
-    return '\n'.join(
-        line for line in text.split('\n')
+    return os.linesep.join(
+        line for line in text.split(os.linesep)
         if ('SNIMissingWarning' not in line) and ('InsecurePlatformWarning' not in line)
     )
 
@@ -134,7 +136,7 @@ def run_command(path_to_script, args):
         stdin=subprocess.PIPE,
         stdout=subprocess.PIPE,
         stderr=subprocess.PIPE,
-        close_fds=True
+        close_fds=platform.system() != 'Windows'
     )
     p.stdin.close()
     reader1 = threading.Thread(target=stdout.read_from, args=[p.stdout])
@@ -156,7 +158,7 @@ def print_text_indented(text):
     """
     Prints text that may include weird characters, indented four spaces.
     """
-    for line in text.split('\n'):
+    for line in text.split(os.linesep):
         print('   ', repr(line)[1:-1])
 
 
@@ -190,6 +192,14 @@ class CommandLine(object):
     def __init__(self, path_to_script):
         self.path_to_script = path_to_script
 
+    def run_command(self, args):
+        """
+        Runs the command with the given arguments, returns a tuple in form of
+        (succeeded, stdout)
+        """
+        status, stdout, stderr = run_command(self.path_to_script, args)
+        return status == 0 and stderr == '', stdout
+
     def should_succeed(self, args, expected_pattern=None):
         """
         Runs the command-line with the given arguments.  Raises an exception
@@ -202,7 +212,7 @@ class CommandLine(object):
             sys.exit(1)
         if stderr != '':
             failed = False
-            for line in (s.strip() for s in stderr.split('\n')):
+            for line in (s.strip() for s in stderr.split(os.linesep)):
                 if not any(p.match(line) for p in self.EXPECTED_STDERR_PATTERNS):
                     print('Unexpected stderr line:', repr(line))
                     failed = True
@@ -235,6 +245,8 @@ class CommandLine(object):
             print('ERROR: should have failed')
             sys.exit(1)
         if re.search(expected_pattern, stdout + stderr) is None:
+            print(expected_pattern)
+            print(stdout + stderr)
             error_and_exit('did not match pattern: ' + expected_pattern)
 
     def list_file_versions(self, bucket_name):
@@ -281,7 +293,7 @@ def clean_buckets(b2_tool, bucket_name_prefix):
     text = b2_tool.should_succeed(['list_buckets'])
 
     buckets = {}
-    for line in text.split('\n')[:-1]:
+    for line in text.split(os.linesep)[:-1]:
         words = line.split()
         if len(words) != 3:
             error_and_exit('bad list_buckets line: ' + line)
@@ -327,6 +339,29 @@ def tearDown_envvar_test(envvar_name):
         del os.environ[envvar_name]
 
 
+def download_test(b2_tool, bucket_name):
+
+    file_to_upload = 'README.md'
+
+    uploaded_a = b2_tool.should_succeed_json(
+        ['upload_file', '--noProgress', '--quiet', bucket_name, file_to_upload, 'a']
+    )
+    with TempDir() as dir_path:
+        p = lambda fname: os.path.join(dir_path, fname)
+        b2_tool.should_succeed(['download_file_by_name', '--noProgress', bucket_name, 'a', p('a')])
+        assert read_file(p('a')) == read_file(file_to_upload)
+        b2_tool.should_succeed(
+            ['download_file_by_id', '--noProgress', uploaded_a['fileId'],
+             p('b')]
+        )
+        assert read_file(p('b')) == read_file(file_to_upload)
+
+    # there is just one file, so clean after itself for faster execution
+    b2_tool.should_succeed(['delete_file_version', uploaded_a['fileName'], uploaded_a['fileId']])
+    b2_tool.should_succeed(['delete_bucket', bucket_name])
+    return True
+
+
 def basic_test(b2_tool, bucket_name):
 
     file_to_upload = 'README.md'
@@ -334,7 +369,7 @@ def basic_test(b2_tool, bucket_name):
 
     hex_sha1 = hashlib.sha1(read_file(file_to_upload)).hexdigest()
 
-    uploaded_a = b2_tool.should_succeed_json(
+    b2_tool.should_succeed(
         ['upload_file', '--noProgress', '--quiet', bucket_name, file_to_upload, 'a']
     )
     b2_tool.should_succeed(['upload_file', '--noProgress', bucket_name, file_to_upload, 'a'])
@@ -360,10 +395,7 @@ def basic_test(b2_tool, bucket_name):
     )
 
     b2_tool.should_succeed(
-        ['download_file_by_name', '--noProgress', bucket_name, 'b/1', '/dev/null']
-    )
-    b2_tool.should_succeed(
-        ['download_file_by_id', '--noProgress', uploaded_a['fileId'], '/dev/null']
+        ['download_file_by_name', '--noProgress', bucket_name, 'b/1', os.devnull]
     )
 
     b2_tool.should_succeed(['hide_file', bucket_name, 'c'])
@@ -396,13 +428,16 @@ def basic_test(b2_tool, bucket_name):
     )
     should_equal(['c'], [f['fileName'] for f in list_of_files['files']])
 
-    b2_tool.should_succeed(['ls', bucket_name], r'^a\nb/\nd\n')
+    b2_tool.should_succeed(['ls', bucket_name], '^a{0}b/{0}d{0}'.format(os.linesep))
     b2_tool.should_succeed(
-        ['ls', '--long', bucket_name], r'^4_z.*upload.*a\n.*-.*b/\n4_z.*upload.*d\n'
+        ['ls', '--long', bucket_name],
+        '^4_z.*upload.*a{0}.*-.*b/{0}4_z.*upload.*d{0}'.format(os.linesep)
     )
-    b2_tool.should_succeed(['ls', '--versions', bucket_name], r'^a\na\nb/\nc\nc\nd\n')
-    b2_tool.should_succeed(['ls', bucket_name, 'b'], r'^b/1\nb/2\n')
-    b2_tool.should_succeed(['ls', bucket_name, 'b/'], r'^b/1\nb/2\n')
+    b2_tool.should_succeed(
+        ['ls', '--versions', bucket_name], '^a{0}a{0}b/{0}c{0}c{0}d{0}'.format(os.linesep)
+    )
+    b2_tool.should_succeed(['ls', bucket_name, 'b'], '^b/1{0}b/2{0}'.format(os.linesep))
+    b2_tool.should_succeed(['ls', bucket_name, 'b/'], '^b/1{0}b/2{0}'.format(os.linesep))
 
     file_info = b2_tool.should_succeed_json(['get_file_info', second_c_version['fileId']])
     expected_info = {
@@ -413,17 +448,74 @@ def basic_test(b2_tool, bucket_name):
     should_equal(expected_info, file_info['fileInfo'])
 
     b2_tool.should_succeed(['delete_file_version', 'c', first_c_version['fileId']])
-    b2_tool.should_succeed(['ls', bucket_name], r'^a\nb/\nc\nd\n')
+    b2_tool.should_succeed(['ls', bucket_name], '^a{0}b/{0}c{0}d{0}'.format(os.linesep))
 
     b2_tool.should_succeed(['make_url', second_c_version['fileId']])
 
-    new_creds = '/tmp/b2_account_info'
+
+def key_restrictions_test(b2_tool, bucket_name):
+
+    second_bucket_name = 'test-b2-command-line-' + random_hex(8)
+    b2_tool.should_succeed(['create-bucket', second_bucket_name, 'allPublic'],)
+
+    key_one_name = 'clt-testKey-01' + random_hex(6)
+    created_key_stdout = b2_tool.should_succeed(
+        [
+            'create-key',
+            key_one_name,
+            'listFiles,listBuckets,readFiles,writeKeys',
+        ]
+    )
+    key_one_id, key_one = created_key_stdout.split()
+
+    b2_tool.should_succeed(['authorize_account', key_one_id, key_one],)
+
+    b2_tool.should_succeed(['get-bucket', bucket_name],)
+    b2_tool.should_succeed(['get-bucket', second_bucket_name],)
+
+    key_two_name = 'clt-testKey-02' + random_hex(6)
+    created_key_two_stdout = b2_tool.should_succeed(
+        [
+            'create-key',
+            '--bucket',
+            bucket_name,
+            key_two_name,
+            'listFiles,listBuckets,readFiles',
+        ]
+    )
+    key_two_id, key_two = created_key_two_stdout.split()
+
+    b2_tool.should_succeed(['authorize_account', key_two_id, key_two],)
+    b2_tool.should_succeed(['get-bucket', bucket_name],)
+    b2_tool.should_succeed(['list-file-names', bucket_name],)
+
+    failed_bucket_err = r'ERROR: Application key is restricted to bucket: ' + bucket_name
+    b2_tool.should_fail(['get-bucket', second_bucket_name], failed_bucket_err)
+
+    failed_list_files_err = r'ERROR: Application key is restricted to bucket: ' + bucket_name
+    b2_tool.should_fail(['list-file-names', second_bucket_name], failed_list_files_err)
+
+    # reauthorize with more capabilities for clean up
+    b2_tool.should_succeed(['authorize_account', sys.argv[1], sys.argv[2]])
+    b2_tool.should_succeed(['delete-bucket', second_bucket_name])
+    b2_tool.should_succeed(['delete-key', key_one_id])
+    b2_tool.should_succeed(['delete-key', key_two_id])
+
+
+def account_test(b2_tool, bucket_name):
+    # actually a high level operations test - we run bucket tests here since this test doesn't use it
+    b2_tool.should_succeed(['delete_bucket', bucket_name])
+    new_bucket_name = bucket_name[:-8] + random_hex(
+        8
+    )  # apparently server behaves erratically when we delete a bucket and recreate it right away
+    b2_tool.should_succeed(['create_bucket', new_bucket_name, 'allPrivate'])
+    b2_tool.should_succeed(['update_bucket', new_bucket_name, 'allPublic'])
+
+    new_creds = os.path.join(tempfile.gettempdir(), 'b2_account_info')
     setup_envvar_test('B2_ACCOUNT_INFO', new_creds)
     b2_tool.should_succeed(['clear_account'])
     bad_application_key = sys.argv[2][:-8] + ''.join(reversed(sys.argv[2][-8:]))
-    b2_tool.should_fail(
-        ['authorize_account', sys.argv[1], bad_application_key], r'nvalid authorization'
-    )
+    b2_tool.should_fail(['authorize_account', sys.argv[1], bad_application_key], r'unauthorized')
     b2_tool.should_succeed(['authorize_account', sys.argv[1], sys.argv[2]])
     tearDown_envvar_test('B2_ACCOUNT_INFO')
 
@@ -467,18 +559,6 @@ def _sync_test_using_dir(b2_tool, bucket_name, dir_):
 
         p = lambda fname: os.path.join(dir_path, fname)
 
-        #dir_path is empty so sync should fail unless --allowEmptySource is specified
-        if VERSION_0_COMPATIBILITY:
-            b2_tool.should_succeed(['sync', '--noProgress', dir_path, b2_sync_point])
-        else:
-            b2_tool.should_fail(
-                ['sync', '--noProgress', dir_path, b2_sync_point],
-                r'Exception: Directory .* is empty'
-            )
-
-        b2_tool.should_succeed(
-            ['sync', '--noProgress', '--allowEmptySource', dir_path, b2_sync_point]
-        )
         file_versions = b2_tool.list_file_versions(bucket_name)
         should_equal([], file_version_summary(file_versions))
 
@@ -496,7 +576,7 @@ def _sync_test_using_dir(b2_tool, bucket_name, dir_):
         # now upload
         b2_tool.should_succeed(
             ['sync', '--noProgress', dir_path, b2_sync_point],
-            expected_pattern="/d could not be accessed"
+            expected_pattern="d could not be accessed"
         )
         file_versions = b2_tool.list_file_versions(bucket_name)
         should_equal(
@@ -584,7 +664,7 @@ def _sync_test_using_dir(b2_tool, bucket_name, dir_):
             ], file_version_summary(file_versions)
         )
 
-        #should new version of c
+        #should upload new version of c
         b2_tool.should_succeed(
             [
                 'sync', '--noProgress', '--keepDays', '10', '--compareVersions', 'modTime',
@@ -633,6 +713,32 @@ def sync_down_helper(b2_tool, bucket_name, folder_in_bucket):
         should_equal(['a', 'b'], sorted(os.listdir(local_path)))
 
 
+def sync_long_path_test(b2_tool, bucket_name):
+    """
+    test sync with very long path (overcome windows 260 character limit)
+    """
+    b2_sync_point = 'b2://' + bucket_name
+
+    long_path = '/'.join(
+        (
+            'extremely_long_path_which_exceeds_windows_unfortunate_260_character_path_limit',
+            'and_needs_special_prefixes_containing_backslashes_added_to_overcome_this_limitation',
+            'when_doing_so_beware_leaning_toothpick_syndrome_as_it_can_cause_frustration',
+            'see_also_xkcd_1638'
+        )
+    )
+
+    with TempDir() as dir_path:
+        local_long_path = os.path.normpath(os.path.join(dir_path, long_path))
+        fixed_local_long_path = fix_windows_path_limit(local_long_path)
+        os.makedirs(os.path.dirname(fixed_local_long_path))
+        write_file(fixed_local_long_path, b'asdf')
+
+        b2_tool.should_succeed(['sync', '--noProgress', '--delete', dir_path, b2_sync_point])
+        file_versions = b2_tool.list_file_versions(bucket_name)
+        should_equal(['+ ' + long_path], file_version_summary(file_versions))
+
+
 def main():
 
     if len(sys.argv) < 3:
@@ -640,19 +746,24 @@ def main():
     path_to_script = 'b2'
     account_id = sys.argv[1]
     application_key = sys.argv[2]
+    defer_cleanup = True
 
     test_map = {
+        'account': account_test,
         'basic': basic_test,
+        'keys': key_restrictions_test,
         'sync_down': sync_down_test,
         'sync_up': sync_up_test,
         'sync_up_no_prefix': sync_test_no_prefix,
+        'sync_long_path': sync_long_path_test,
+        'download': download_test,
     }
 
     if len(sys.argv) >= 4:
         tests_to_run = sys.argv[3:]
         for test_name in tests_to_run:
             if test_name not in test_map:
-                error_and_exit('unknown test: ' + test_name)
+                error_and_exit('unknown test: "%s"' % (test_name,))
     else:
         tests_to_run = sorted(six.iterkeys(test_map))
 
@@ -661,6 +772,7 @@ def main():
 
     b2_tool = CommandLine(path_to_script)
 
+    global_dirty = False
     # Run each of the tests in its own empty bucket
     for test_name in tests_to_run:
 
@@ -671,18 +783,17 @@ def main():
 
         b2_tool.should_succeed(['clear_account'])
 
-        bad_application_key = application_key[:-8] + ''.join(reversed(application_key[-8:]))
-        b2_tool.should_fail(
-            ['authorize_account', account_id, bad_application_key], r'Invalid authorization'
-        )
         b2_tool.should_succeed(['authorize_account', account_id, application_key])
 
         bucket_name_prefix = 'test-b2-command-line-' + account_id
-        clean_buckets(b2_tool, bucket_name_prefix)
+        if not defer_cleanup:
+            clean_buckets(b2_tool, bucket_name_prefix)
         bucket_name = bucket_name_prefix + '-' + random_hex(8)
 
-        b2_tool.should_succeed(['create_bucket', bucket_name, 'allPrivate'])
-        b2_tool.should_succeed(['update_bucket', bucket_name, 'allPublic'])
+        success, _ = b2_tool.run_command(['create_bucket', bucket_name, 'allPublic'])
+        if not success:
+            clean_buckets(b2_tool, bucket_name_prefix)
+            b2_tool.should_succeed(['create_bucket', bucket_name, 'allPublic'])
 
         print('#')
         print('# Running test:', test_name)
@@ -690,8 +801,17 @@ def main():
         print()
 
         test_fcn = test_map[test_name]
-        test_fcn(b2_tool, bucket_name)
+        dirty = not test_fcn(b2_tool, bucket_name)
+        global_dirty = global_dirty or dirty
 
+    if global_dirty:
+        print('#' * 70)
+        print('#')
+        print('# The last test was run, cleaning up')
+        print('#')
+        print('#' * 70)
+        print()
+        clean_buckets(b2_tool, bucket_name_prefix)
     print()
     print("ALL OK")
 
